__author__ = 'Michele Bannister   git:@mtbannister'

import numpy as np
import matplotlib
import matplotlib.pyplot as plt

# from mpld3 import plugins, utils
import brewer2mpl

ALMOST_BLACK = '#262626'

# nicer defaults
# every colourbrewer scale: http://bl.ocks.org/mbostock/5577023
# hues create primary visual differences between classes; does not imply magnitude differences between classes
set2 = brewer2mpl.get_map('Set2', 'qualitative', 8).mpl_colors
# # sequential for ordered data where light is low values, dark is high values
# ylorrd = brewer2mpl.get_map('YlOrRd', 'sequential', 9).mpl_colors
# # diverging for equal emphasis on mid-range critical values and extremes at both ends of data range
# puor = brewer2mpl.get_map('PuOr', 'diverging', 11).mpl_colors

# FIXME: figure out how to export these to a figure

# rcParams['figure.figsize'] = (10, 10)
# rcParams['figure.dpi'] = 150
# rcParams['axes.color_cycle'] = puor
matplotlib.rcParams['font.size'] = 12  # good for posters/slides
matplotlib.rcParams['patch.facecolor'] = set2[0]
matplotlib.rcParams['font.family'] = 'sans-serif'
matplotlib.rcParams['font.sans-serif'] = ['Sofia Pro', 'Tahoma']


def remove_border(axes=None, keep=('left', 'bottom'), remove=('right', 'top'), labelcol=ALMOST_BLACK):
    """
    Minimize chart junk by stripping out unnecessary plot borders and axis ticks.
    The top/right/left/bottom keywords toggle whether the corresponding plot border is drawn
    """
    ax = axes or plt.gca()
    for spine in remove:
        ax.spines[spine].set_visible(False)
    for spine in keep:
        ax.spines[spine].set_linewidth(0.5)
        # ax.spines[spine].set_color('white')

    # remove all ticks, then add back the ones in keep
    # Does this also need to specify the ticks' colour, given the axes/labels are changed?
    ax.yaxis.set_ticks_position('none')
    ax.xaxis.set_ticks_position('none')
    # ax.xaxis.set_ticklabels("")
    # ax.yaxis.set_ticklabels("")

    for spine in keep:
        if spine == 'top':
            ax.xaxis.tick_top()
        if spine == 'bottom':
            ax.xaxis.tick_bottom()
            # match the label colour to that of the axes
            ax.xaxis.label.set_color(labelcol)
            ax.xaxis.set_tick_params(color=labelcol, labelcolor=labelcol)
        if spine == 'left':
            ax.yaxis.tick_left()
            ax.yaxis.label.set_color(labelcol)
            ax.yaxis.set_tick_params(color=labelcol, labelcolor=labelcol)
        if spine == 'right':
            ax.yaxis.tick_right()

    return


def clean_legend():
    # as per prettyplotlib.
    # Remove the line around the legend box, and instead fill it with a light grey
    # Also only use one point for the scatterplot legend because the user will
    # get the idea after just one, they don't need three.
    light_grey = np.array([float(248) / float(255)] * 3)
    legend = ax.legend(frameon=True, scatterpoints=1)
    rect = legend.get_frame()
    rect.set_facecolor(light_grey)
    rect.set_linewidth(0.0)

    # Change the legend label colors to almost black, too
    texts = legend.texts
    for t in texts:
        t.set_color(almost_black)


# class LinkedBrush(plugins.PluginBase):
#     JAVASCRIPT = r"""
#     var LinkedBrushPlugin = function(fig, prop){
#       this.fig = fig;
#       this.prop = mpld3.process_props(this, prop, {}, ["id"]);
#
#       var brush_plug = this;
#
#       mpld3.Toolbar.prototype.buttonDict["brush"] = mpld3.ButtonFactory({
#         onClick: this.onClick.bind(this),
#         draw: function(){
#             mpld3.BaseButton.prototype.draw.apply(this);
#             var enable_zoom = brush_plug.fig.enable_zoom.bind(brush_plug.fig);
#             var disable_brush = brush_plug.disable.bind(brush_plug);
#             brush_plug.fig.enable_zoom = function(){
#                    disable_brush();
#                    fig.toolbar.toolbar.selectAll(".mpld3-brushbutton")
#                        .classed({pressed: false,
#                                  active: false});
#                    enable_zoom();
#             };
#             this.onClick();  // enable the button
#         },
#         icon: "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABmJLR0QA/wD/AP+gvaeTAAAACXBI
#         \nWXMAAEQkAABEJAFAZ8RUAAAAB3RJTUUH3gMCEiQKB9YaAgAAAWtJREFUOMuN0r1qVVEQhuFn700k\nnfEvBq0iNiIiOKXgH4KCaBeIhWARK
#         /EibLwFCwVLjyAWaQzRGG9grC3URkHUBKKgRuWohWvL5pjj\nyTSLxcz7rZlZHyMiItqzFxGTEVF18/UoODNFxDIO4x12dkXqTcBPsCUzD
#         +AK3ndFqhHwEsYz82gn\nN4dbmMRK9R/4KY7jAvbiWmYeHBT5Z4QCP8J1rGAeN3GvU3Mbl/Gq3qCDcxjLzOV+v78fq/iFIxFx
#         \nPyJ2lNJpfBy2g59YzMyzEbEVLzGBJjOriLiBq5gaJrCIU3hcRCbwAtuwjm/Yg/V6I9NgDA1OR8RC\nZq6Vcd7iUwtn5h8fdMBdETGPE
#         +Xe4ExELDRNs4bX2NfCUHe+7UExyfkCP8MhzOA7PuAkvrbwXyNF
#         \nxF3MDqxiqlhXC7SPdaOKiN14g0u4g3H0MvOiTUSNY3iemb0ywmfMdfYyUmAJ2yPiBx6Wr/oy2Oqw\n+A1SupBzAOuE
#         /AAAAABJRU5ErkJggg==\n",
#        });
#     }
#
#     LinkedBrushPlugin.prototype.onClick = function(){
#       if(this.enabled){
#         this.disable();
#       }else{
#         this.enable();
#         this.fig.disable_zoom();
#         this.fig.toolbar.toolbar.selectAll(".mpld3-movebutton")
#                    .classed({pressed: false,
#                              active: false});
#       }
#       this.fig.toolbar.toolbar.selectAll(".mpld3-brushbutton")
#                 .classed({pressed: this.enabled,
#                           active: !this.enabled});
#     }
#
#     LinkedBrushPlugin.prototype.draw = function(){
#       var obj = mpld3.get_element(this.prop.id);
#       var fig = this.fig;
#       var dataKey = ("offsets" in obj.prop) ? "offsets" : "data";
#
#       mpld3.insert_css("#" + fig.figid + " rect.extent",
#                        {"fill": "#000",
#                         "fill-opacity": .125,
#                         "stroke": "#fff"});
#
#       mpld3.insert_css("#" + fig.figid + " path.mpld3-hidden",
#                        {"stroke": "#ccc !important",
#                         "fill": "#ccc !important"});
#
#       var dataClass = "mpld3data-" + obj.prop[dataKey];
#
#       var brush = d3.svg.brush()
#                       .on("brushstart", brushstart)
#                       .on("brush", brushmove)
#                       .on("brushend", brushend);
#
#       // Label all data points for access below
#       fig.axes.forEach(function(ax){
#          ax.elements.forEach(function(el){
#             if(el.prop[dataKey] === obj.prop[dataKey]){
#                el.group.classed(dataClass, true);
#             }
#          });
#          brush.x(ax.x).y(ax.y);
#          //ax.axes.call(brush);
#       });
#
#
#       // For fast brushing, precompute a list of selection properties
#       // properties to apply to the selction.
#       var data_map = [];
#       var dataToBrush = fig.canvas.selectAll("." + dataClass)
#                            .each(function(){
#                               for(var i=0; i<fig.axes.length; i++){
#                                 var ax = fig.axes[i];
#                                 for(var j=0; j<ax.elements.length; j++){
#                                   var el = ax.elements[j];
#                                   if("group" in el && el.group[0][0] === this){
#                                     data_map.push({i_ax: i,
#                                                    ix: el.prop.xindex,
#                                                    iy: el.prop.yindex});
#                                     return;
#                                   }
#                                 }
#                               }
#                             });
#
#       dataToBrush.data(data_map)
#                  .call(brush);
#
#       var currentData;
#
#       function brushstart(d){
#         if(currentData != this){
#           d3.select(currentData).call(brush.clear());
#           currentData = this;
#           brush.x(fig.axes[d.i_ax].x)
#                .y(fig.axes[d.i_ax].y);
#         }
#       }
#
#       function brushmove(d){
#         var e = brush.extent();
#         dataToBrush.selectAll("path")
#                    .classed("mpld3-hidden",
#                        function(p) {
#                            return e[0][0] > p[d.ix] ||  e[1][0] < p[d.ix] ||
#                                   e[0][1] > p[d.iy] || e[1][1] < p[d.iy];
#                        });
#       }
#
#       function brushend(d){
#         if (brush.empty()){
#             dataToBrush.selectAll("path")
#                        .classed("mpld3-hidden", false);
#         }
#       }
#
#       function brushend_clear(){
#         d3.select(this).call(brush.clear());
#       }
#
#       this.enable = function(){
#         brush.on("brushstart", brushstart)
#              .on("brush", brushmove)
#              .on("brushend", brushend);
#         this.fig.canvas.selectAll("rect.background")
#               .style("cursor", "crosshair");
#         this.fig.canvas.selectAll("rect.extent, rect.resize")
#               .style("display", null);
#         this.enabled = true;
#       }
#
#       this.disable = function(){
#         brush.on("brushstart", null)
#              .on("brush", null)
#              .on("brushend", brushend_clear)
#              .clear();
#         this.fig.canvas.selectAll("rect.background")
#               .style("cursor", null);
#         this.fig.canvas.selectAll("rect.extent, rect.resize")
#               .style("display", "none");
#         this.enabled = false;
#       }
#
#       this.disable();
#     }
#
#     mpld3.register_plugin("linkedbrush", LinkedBrushPlugin);
#     """
#
#     def __init__(self, points):
#         if isinstance(points, matplotlib.lines.Line2D):
#             suffix = "pts"
#         else:
#             suffix = None
#
#         self.dict_ = {"type": "linkedbrush",
#                       "clear_toolbar": False,
#                       "id": utils.get_id(points, suffix),
#                       "buttons": "brush"}



from unittest import TestCase
from astropy import units

from ossos import cameras


class TestCamera(TestCase):

    def setUp(self):
        self.camera = cameras.Camera(0*units.degree, 0*units.degree)

    def test_offset(self):
        self.camera.offset(index=22)
        self.assertEqual(self.camera.ra, 0*units.degree)
        self.assertEqual(self.camera.dec, 0*units.degree)
        self.fail()

    def test_coordinate(self):
        self.fail()

    def test_coord(self):
        self.fail()

    def test_ra(self):
        self.fail()

    def test_dec(self):
        self.fail()

    def test_origin(self):
        self.fail()

    def test_origin(self):
        self.fail()

    def test_set_origin(self):
        self.fail()

    def test_polygon(self):
        self.fail()

    def test_geometry(self):
        self.fail()

    def test_separation(self):
        self.fail()



###############################################################################
#                                                                            ##
# Copyright 2013 by its authors                                              ##
# See COPYING, AUTHORS                                                       ##
#                                                                            ##
# This file is part of OSSOS Moving Object Pipeline (OSSOS-MOP)              ##
#                                                                            ##
#    OSSOS-MOP is free software: you can redistribute it and/or modify       ##
#    it under the terms of the GNU General Public License as published by    ##
#    the Free Software Foundation, either version 3 of the License, or       ##
#    (at your option) any later version.                                     ##
#                                                                            ##
#    OSSOS-MOP is distributed in the hope that it will be useful,            ##
#    but WITHOUT ANY WARRANTY; without even the implied warranty of          ##
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           ##
#    GNU General Public License for more details.                            ##
#                                                                            ##
#    You should have received a copy of the GNU General Public License       ##
#    along with OSSOS-MOP.  If not, see <http://www.gnu.org/licenses/>.      ##
#                                                                            ##
###############################################################################
"""step1 is to run the two source finding algorithms in the image.

step1jmp is a stand-alone fortran code from Jean-Marc Petit et al.
step1matt is a script from M. Holman that runs E. Bertain's sExtractor.

"""

import argparse
from astropy.io import fits
import logging
import os
from ossos import storage
from ossos import util
import sys

_SEX_THRESHOLD = 1.1
_WAVE_THRESHOLD = 2.7
_FWHM = 4.0
_MAX_COUNT = 30000

dependency = "mkpsf"
task = "step1"


def _get_weight_map(filename, ccd):
    # for step1matt we need the weight image
    hdulist = fits.open(filename)
    flat_name = hdulist[0].header.get('FLAT', 'weight.fits')
    parts = os.path.splitext(flat_name)
    if parts[1] == '.fz':
        flat_name = os.path.splitext(parts[0])[0]
    else:
        flat_name = parts[0]

    logging.info("Getting the Flat Field image.")
    try:
        flat_filename = storage.get_image(flat_name, ccd, version='', ext='fits', subdir='calibrators')
    except Exception:
        flat_filename = storage.get_image(flat_name, ccd, version='', ext='fits', subdir='old_calibrators')

    if os.access('weight.fits', os.R_OK):
        os.unlink('weight.fits')

    if not os.access('weight.fits', os.R_OK):
        os.symlink(flat_filename, 'weight.fits')

    return flat_filename


def run(expnum,
        ccd,
        prefix='',
        version='p',
        sex_thresh=_SEX_THRESHOLD,
        wave_thresh=_WAVE_THRESHOLD,
        maxcount=_MAX_COUNT,
        dry_run=False,
        force=True):
    """run the actual step1jmp/matt codes.

    expnum: the CFHT expousre to process
    ccd: which ccd in the mosaic to process
    fwhm: the image quality, FWHM, of the image.  In pixels.
    sex_thresh: the detection threhold to run sExtractor at
    wave_thresh: the detection threshold for wavelet
    maxcount: saturation level

    """
    message = storage.SUCCESS

    if storage.get_status(task, prefix, expnum, version, ccd) and not force:
        logging.info("{} completed successfully for {} {} {} {}".format(task, prefix, expnum, version, ccd))
        return

    with storage.LoggingManager(task, prefix, expnum, ccd, version, dry_run):
        try:
            if not storage.get_status(dependency, prefix, expnum, version, ccd):
                raise IOError(35, "Cannot start {} as {} not yet completed for {}{}{}{:02d}".format(
                    task, dependency, prefix, expnum, version, ccd))
            logging.info("Retrieving imaging and input parameters from VOSpace")
            storage.get_file(expnum, ccd, prefix=prefix, version=version, ext='mopheader')
            filename = storage.get_image(expnum, ccd, version=version, prefix=prefix)
            fwhm = storage.get_fwhm(expnum, ccd, prefix=prefix, version=version)
            basename = os.path.splitext(filename)[0]

            _get_weight_map(filename, ccd)

            logging.info("Launching step1jmp")
            logging.info(util.exec_prog(['step1jmp',
                                         '-f', basename,
                                         '-t', str(wave_thresh),
                                         '-w', str(fwhm),
                                         '-m', str(maxcount)]))

            logging.info(util.exec_prog(['step1matt',
                                         '-f', basename,
                                         '-t', str(sex_thresh),
                                         '-w', str(fwhm),
                                         '-m', str(maxcount)]))

            if os.access('weight.fits', os.R_OK):
                os.unlink('weight.fits')

            if not dry_run:
                for ext in ['obj.jmp', 'obj.matt']:
                    obj_uri = storage.get_uri(expnum, ccd, version=version, ext=ext,
                                              prefix=prefix)
                    obj_filename = basename + "." + ext
                    count = 0
                    with open(obj_filename, 'r'):
                        while True:
                            try:
                                count += 1
                                logging.info("Attempt {} to copy {} -> {}".format(count, obj_filename, obj_uri))
                                storage.copy(obj_filename, obj_uri)
                                break
                            except Exception as ex:
                                if count > 10:
                                    raise ex
            logging.info(message)
        except Exception as ex:
            message = str(ex)
            logging.error(message)

        if not dry_run:
            storage.set_status(task, prefix, expnum, version, ccd, status=message)


def main():
    # Must be running as a script

    parser = argparse.ArgumentParser(
        description='Run step1jmp and step1matt on a given exposure.')

    parser.add_argument("--ccd", "-c",
                        action="store",
                        default=None,
                        type=int,
                        dest="ccd")
    parser.add_argument("--ignore", help="Try to run even in previous step failed.",
                        default=False,
                        action="store_true")
    parser.add_argument("--fk", help="add the fk prefix on processing?",
                        default=False,
                        action='store_true')
    parser.add_argument("--dbimages",
                        action="store",
                        default="vos:OSSOS/dbimages",
                        help='vospace dbimages containerNode')
    parser.add_argument("--sex_thresh",
                        action="store",
                        type=float,
                        default=_SEX_THRESHOLD,
                        help="sExtractor detection threhold")
    parser.add_argument("--wavelet_thresh",
                        type=float,
                        default=_WAVE_THRESHOLD,
                        help="Wavelet detection threhold")
    parser.add_argument("expnum",
                        type=int,
                        nargs='+',
                        help="expnum(s) to process")
    parser.add_argument("--version",
                        action='version',
                        version='%(prog)s 1.0')
    parser.add_argument('--type', default='p',
                        choices=['o', 'p', 's'], help="which type of image")
    parser.add_argument('--log', default=None, help="Write standard out to this file")
    parser.add_argument("--verbose", "-v",
                        action="store_true")
    parser.add_argument("--debug", '-d',
                        action='store_true')
    parser.add_argument("--force", action="store_true")
    parser.add_argument("--dry-run", action="store_true", help="Do a dry run, not changes to vospce, implies --force")

    cmd_line = " ".join(sys.argv)
    args = parser.parse_args()

    util.set_logger(args)
    logging.info("Starting {}".format(cmd_line))

    storage.DBIMAGES = args.dbimages

    prefix = (args.fk and 'fk') or ''
    version = args.type

    for expnum in args.expnum:
        if args.ccd is None:
            if int(expnum) < 1785619:
                # Last exposures with 36 CCD Megaprime
                ccdlist = list(range(0, 36))
            else:
                # First exposrues with 40 CCD Megaprime
                ccdlist = list(range(0, 40))
        else:
            ccdlist = [args.ccd]
        for ccd in ccdlist:
            run(expnum,
                ccd,
                prefix=prefix,
                version=version,
                sex_thresh=args.sex_thresh,
                wave_thresh=args.wavelet_thresh,
                dry_run=args.dry_run,
                force=args.force)


if __name__ == '__main__':
    sys.exit(main())



from ossos.ph2 import *

if __name__ == "__main__":

    program = Program(runid="16BE91", pi_login="mwilson")
    program_configuration = ProgramConfiguration()
    program.program_configuration = program_configuration

    constraint = Constraint(Identifier(server_token="C1"))

    instrument_config = InstrumentConfig(Identifier(server_token="I1"))

    moving_target = MovingTarget()
    coordinate = Coordinate(10., 10.)
    ephmeris_point = EphmerisPoint(540000.0, coordinate)
    moving_target.ephmeris_points.append(ephmeris_point)
    target = Target(identifier=Identifier(client_token="target_identifier"),
                    name="target_name",
                    moving_target=moving_target)

    ob = observing_block = ObservingBlock(Identifier(client_token="ob_identifier"),
                                          target=target,
                                          constraint=constraint,
                                          instrument_config=instrument_config)

    og = ObservingGroup(Identifier(client_token="og_identifier"))
    og.observing_block_identifiers.append(ob.identifier)

    program_configuration.observing_blocks.append(ob)
    program_configuration.observing_groups.append(og)
    program_configuration.targets.append(target)
    print((program.to_json()))



################################################################################
#                                                                            ##
# Copyright 2013 by its authors                                              ##
# See COPYING, AUTHORS                                                       ##
#                                                                            ##
# This file is part of OSSOS Moving Object Pipeline (OSSOS-MOP)              ##
#                                                                            ##
#    OSSOS-MOP is free software: you can redistribute it and/or modify       ##
#    it under the terms of the GNU General Public License as published by    ##
#    the Free Software Foundation, either version 3 of the License, or       ##
#    (at your option) any later version.                                     ##
#                                                                            ##
#    OSSOS-MOP is distributed in the hope that it will be useful,            ##
#    but WITHOUT ANY WARRANTY; without even the implied warranty of          ##
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           ##
#    GNU General Public License for more details.                            ##
#                                                                            ##
#    You should have received a copy of the GNU General Public License       ##
#    along with OSSOS-MOP.  If not, see <http://www.gnu.org/licenses/>.      ##
#                                                                            ##
################################################################################
"""
Compare the measured fluxes of planted sources against those returned for by digiphot.
"""
__author__ = 'jjk'

import argparse
import logging
import os
import sys

from cadcutils.exceptions import NotFoundException

from ossos import astrom
from ossos import storage
from ossos import util
from ossos.match import match_planted

logger = logging.getLogger('vos')
logger.setLevel(logging.CRITICAL)
logger.addHandler(logging.StreamHandler())

BRIGHT_LIMIT = 23.0
OBJECT_PLANTED = "Object.planted"
MINIMUM_BRIGHT_DETECTIONS = 5
MINIMUM_BRIGHT_FRACTION = 0.5
task = 'astrom_mag_check'


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('field')
    parser.add_argument('ccd')
    parser.add_argument('--expnum', default=None, help="Which exposure is the lead for this astrom file?")
    parser.add_argument('--astrom-filename', default=None, help="Give the astrom file directly instead of looking-up "
                                                                "using the field/ccd naming scheme.")
    parser.add_argument('--reals', action='store_true', default=False)
    parser.add_argument('--type', choices=['o', 'p', 's'], help="Which type of image.", default='s')
    parser.add_argument('--measure3', default='vos:OSSOS/measure3/2013B-L_redo/')
    parser.add_argument('--dbimages', default=None)
    parser.add_argument('--dry-run', action='store_true', default=False)
    parser.add_argument('--force', action='store_true', default=False)
    parser.add_argument("--fk", action="store_true", default=False, help="Do fakes?")
    parser.add_argument('--object-planted', default=OBJECT_PLANTED,
                        help="Name of file contains list of planted objects.")
    parser.add_argument('--bright-limit', default=BRIGHT_LIMIT,
                        help="Sources brighter than this limit {} are used to diagnose planting issues.".format(
                            BRIGHT_LIMIT))
    parser.add_argument('--minimum-bright-detections', default=MINIMUM_BRIGHT_DETECTIONS, type=int,
                        help="required number of detections with mag brighter than bright-limit.")
    parser.add_argument('--minimum-bright-fraction', default=MINIMUM_BRIGHT_FRACTION,
                        help="minimum fraction of objects above bright limit that should be found.")
    parser.add_argument("--debug",
                        action="store_true")
    parser.add_argument("--verbose", "-v",
                        action="store_true")

    cmd_line = " ".join(sys.argv)
    args = parser.parse_args()

    util.set_logger(args)
    logging.info("Starting {}".format(cmd_line))
    prefix = (args.fk and "fk") or ""
    expnum = args.expnum
    version = args.type
    ext = 'cands'
    if args.reals:
        ext = 'reals'

    if args.dbimages is not None:
        storage.DBIMAGES = args.dbimages
        astrom.DATASET_ROOT = args.dbimages
    storage.MEASURE3 = args.measure3
    astrom_uri = storage.get_cands_uri(args.field,
                                       ccd=args.ccd,
                                       version=args.type,
                                       prefix=prefix,
                                       ext="measure3.{}.astrom".format(ext))

    if args.astrom_filename is None:
        astrom_filename = os.path.basename(astrom_uri)
    else:
        astrom_filename = args.astrom_filename

    if not os.access(astrom_filename, os.F_OK):
        astrom_filename = os.path.dirname(astrom_uri) + "/" + astrom_filename

    # Load the list of astrometric observations that will be looked at.
    fk_candidate_observations = astrom.parse(astrom_filename)
    if args.expnum is None:
        expnum = fk_candidate_observations.observations[0].expnum
    else:
        expnum = args.expnum

    match_filename = os.path.splitext(os.path.basename(astrom_filename))[0] + '.match'

    message = storage.SUCCESS
    with storage.LoggingManager(task, prefix, expnum, args.ccd, version, args.dry_run):
        try:
            logging.info(("Comparing planted and measured magnitudes "
                          "for sources in {} and {}\n".format(args.object_planted, astrom_filename)))
            result = match_planted(fk_candidate_observations,
                                   match_filename=match_filename,
                                   object_planted=args.object_planted,
                                   bright_limit=args.bright_limit,
                                   minimum_bright_detections=args.minimum_bright_detections,
                                   bright_fraction=args.minimum_bright_fraction)
            match_uri = storage.get_cands_uri(args.field,
                                              ccd=args.ccd,
                                              version=args.type,
                                              prefix=prefix,
                                              ext="measure3.{}.match".format(ext), block=args.field)
            if not args.dry_run:
                temp_match_file = os.path.basename(match_uri)
                try:
                    storage.copy(match_uri, temp_match_file)
                except NotFoundException as ex:
                    logging.warning(f'No match file found at {match_uri}, creating one.')
                    logging.debug(f'{ex}')
                    pass
                with open(temp_match_file, 'a') as fout:
                    fout.write(open(match_filename, 'r').read())
                storage.copy(temp_match_file, match_uri)
                uri = os.path.dirname(astrom_uri)
                keys = [storage.tag_uri(os.path.basename(astrom_uri))]
                values = [result]
                storage.set_tags_on_uri(uri, keys, values)
            logger.info(message)
        except Exception as err:
            message = str(err)
            logging.error(message)

        if not args.dry_run:
            storage.set_status(task, prefix, expnum, version='', ccd=args.ccd, status=message)

    return


if __name__ == '__main__':
    main()



#!python
"""
Update the astrometric and photometric measurements of an mpc observation based on the header contents of the
observation.
"""
from copy import deepcopy
import time
from astropy import units
import os
import math
import numpy
import logging
import re
import mp_ephem
from ossos import storage, wcs, astrom
from ossos import orbfit
import argparse

__author__ = 'jjk'

# Maximum allowed change in angle during re-measure
TOLERANCE = 2.00 * units.arcsec


def _flipped_ccd(ccd):
    """
    Is this CCD likely flipped?
    
    The MegaCam imager on CFHT has E/N flipped for some of their CCDs.
    
    @param ccd:
    @return:
    """
    return ccd < 18 or ccd in [36, 37]


def remeasure(mpc_in, reset_pixel_coordinates=True):
    """
    Compute the RA/DEC of the line based on the X/Y in the comment and the WCS of the associated image.

    Comment of supplied astrometric line (mpc_in) must be in OSSOSComment format.

    @param mpc_in: An line of astrometric measurement to recompute the RA/DEC from the X/Y in the comment.
    @type mpc_in: mp_ephem.Observation
    @param reset_pixel_coordinates: try and determine correct X/Y is X/Y doesn't map to correct RA/DEC value
    @type reset_pixel_coordinates: bool
    @type reset_pixecl_coordinates: bool

    """
    if mpc_in.null_observation:
        return mpc_in
    mpc_obs = deepcopy(mpc_in)
    logging.debug("rm start: {}".format(mpc_obs.to_string()))

    if not isinstance(mpc_obs.comment, mp_ephem.ephem.OSSOSComment):
        logging.error("Failed to convert comment line")
        return mpc_in

    parts = re.search('(?P<expnum>\d{7})(?P<type>\S)(?P<ccd>\d\d)', str(mpc_obs.comment.frame))
    if not parts:
        logging.error("Failed to parse expnum from frame info in comment line")
        return mpc_in
    ccd = int(parts.group('ccd'))
    expnum = int(parts.group('expnum'))
    exp_type = parts.group('type')

    try:
        header = _connection_error_wrapper(storage._get_sghead, expnum)[ccd+1]
    except IOError as ioerr:
        logging.error(str(ioerr))
        logging.error("Failed to get astrometric header for: {}".format(mpc_obs))
        return mpc_in

    this_wcs = wcs.WCS(header)

    coordinate = this_wcs.xy2sky(mpc_obs.comment.x, mpc_obs.comment.y, usepv=True)
    mpc_obs.coordinate = coordinate[0].to('degree').value, coordinate[1].to('degree').value
    sep = mpc_in.coordinate.separation(mpc_obs.coordinate)

    if sep > TOLERANCE*20 and mpc_in.discovery and _flipped_ccd(ccd):
        logging.warn("Large ({}) offset using X/Y in comment line to compute RA/DEC".format(sep))
        if reset_pixel_coordinates:
            logging.info("flipping/flopping the discvoery x/y position recorded.")
            x = header['NAXIS1'] - mpc_obs.comment.x + 1
            y = header['NAXIS2'] - mpc_obs.comment.y + 1
            new_coordinate = this_wcs.xy2sky(x, y, usepv=True)
            new_sep = mpc_in.coordinate.separation(new_coordinate)
            if new_sep < TOLERANCE*2:
                mpc_obs.coordinate = new_coordinate
                mpc_obs.comment.x = x
                mpc_obs.comment.y = y
                sep = new_sep

    if sep > TOLERANCE:
        # use the old header RA/DEC to predict the X/Y and then use that X/Y to get new RA/DEC
        logging.warn("sep: {} --> large offset when using comment line X/Y to compute RA/DEC")
        if reset_pixel_coordinates:
           logging.warn("Using RA/DEC and original WCS to compute X/Y and replacing X/Y in comment.".format(sep))
           header2 = _connection_error_wrapper(storage.get_astheader, expnum, ccd)
           image_wcs = wcs.WCS(header2)
           (x, y) = image_wcs.sky2xy(mpc_in.coordinate.ra.degree, mpc_in.coordinate.dec.degree, usepv=False)
           mpc_obs.coordinate = this_wcs.xy2sky(x, y, usepv=True)
           mpc_obs.comment.x = x
           mpc_obs.comment.y = y
           logging.info("Coordinate changed: ({:5.2f},{:5.2f}) --> ({:5.2f},{:5.2f})".format(mpc_obs.comment.x,
                                                                                             mpc_obs.comment.y,
                                                                                             x, y))

    if mpc_obs.comment.mag_uncertainty is not None:
        try:
            merr = float(mpc_obs.comment.mag_uncertainty)
            fwhm = float(_connection_error_wrapper(storage.get_fwhm, expnum, ccd))
            centroid_err = merr * fwhm * header['PIXSCAL1']
            logging.debug("Centroid uncertainty:  {} {} => {}".format(merr, fwhm, centroid_err))
        except Exception as err:
            logging.error(str(err))
            logging.error("Failed to compute centroid_err for observation:\n"
                          "{}\nUsing default of 0.2".format(mpc_obs.to_string()))
            centroid_err = 0.2
    else:
        centroid_err = 0.2

    mpc_obs.comment.astrometric_level = header.get('ASTLEVEL', "0")

    try:
        asterr = float(header['ASTERR'])
        residuals = (asterr ** 2 + centroid_err ** 2) ** 0.5
        logging.debug("Residuals: {} {} => {}".format(asterr, centroid_err, residuals))
    except Exception as err:
        logging.error(str(err))
        logging.error("Failed while trying to compute plate uncertainty for\n{}".format(mpc_obs.to_string()))
        logging.error('Using default of 0.25')
        residuals = 0.25

    mpc_obs.comment.plate_uncertainty = residuals

    logging.debug("sending back: {}".format(mpc_obs.to_string()))

    return mpc_obs


def _connection_error_wrapper(func, *args, **kwargs):
    """
    Wrap a call to func in a try/except that repeats on ConnectionError
    @param func:
    @param args:
    @param kwargs:
    @return:
    """

    counter = 0
    while counter < 5:
        try:
            result = func(*args, **kwargs)
            return result
        except Exception as ex:
            time.sleep(5)
            counter += 1
            logging.warning(str(ex))


def recompute_mag(mpc_in, skip_centroids=False):
    """
    Get the mag of the object given the mp_ephem.ephem.Observation
    """
    # TODO this really shouldn't need to build a 'reading' to get the cutout...

    from ossos.downloads.cutouts import downloader
    dlm = downloader.ImageCutoutDownloader()

    mpc_obs = deepcopy(mpc_in)
    assert isinstance(mpc_obs, mp_ephem.ephem.Observation)
    assert isinstance(mpc_obs.comment, mp_ephem.ephem.OSSOSComment)

    if mpc_obs.null_observation:
        return mpc_obs

    parts = re.search('(?P<expnum>\d{7})(?P<type>\S)(?P<ccd>\d\d)', mpc_obs.comment.frame)
    if parts is None:
        return mpc_obs
    expnum = parts.group('expnum')
    ccd = parts.group('ccd')
    file_type = parts.group('type')

    observation = astrom.Observation(expnum, file_type, ccd)
    assert isinstance(observation, astrom.Observation)

    ast_header = _connection_error_wrapper(storage._get_sghead, int(expnum))[int(ccd)+1]

    filter_value = None
    for keyword in ['FILTER', 'FILT1 NAME']:
       filter_value = ast_header.get(keyword, None)
       if filter_value is not None:
         if filter_value.startswith('gri'):
             filter_value = 'w'
         else:
             filter_value = filter_value[0]
         break
    # The ZP for the current astrometric lines is the pipeline one.  The new ZP is in the astheader file.
    new_zp = ast_header.get('PHOTZP')

    # The .zeropoint.used value is likely the one used for the original photometry.
    old_zp = _connection_error_wrapper(storage.get_zeropoint, int(expnum), int(ccd))

    reading = astrom.SourceReading(float(mpc_obs.comment.x), float(mpc_obs.comment.y), float(mpc_obs.comment.x),
                                   float(mpc_obs.comment.y), mpc_obs.coordinate.ra.degree,
                                   mpc_obs.coordinate.dec.degree, float(mpc_obs.comment.x), float(mpc_obs.comment.y),
                                   observation, ssos=True, from_input_file=True, null_observation=False,
                                   discovery=mpc_obs.discovery)

    cutout = _connection_error_wrapper(dlm.download_cutout, reading, needs_apcor=True)
    cutout._zmag = new_zp

    if math.fabs(cutout.zmag - old_zp) > 0.3:
        logging.warn("Large change in zeropoint detected: {}  -> {}".format(old_zp, cutout.zmag))

    try:
        PHOT = cutout.get_observed_magnitude(centroid=not skip_centroids and mpc_obs.note1 != "H")
        x = PHOT['XCENTER']
        y = PHOT['YCENTER']
        mag = PHOT['MAG']
        merr = PHOT['MERR']
        cutout.update_pixel_location((x, y), hdu_index=cutout.extno)
        x, y = cutout.observed_source_point
    except Exception as ex:
        logging.error("ERROR: {}".format(str(ex)))
        return mpc_obs

    try:
        if mpc_obs.comment.mag_uncertainty is not None and mpc_obs.comment.mag is not None and math.fabs(mpc_obs.comment.mag - mag) > 3.5 * mpc_obs.comment.mag_uncertainty:
           logging.warn("recomputed magnitude shift large: {} --> {}".format(mpc_obs.mag, mag[0]))
        if math.sqrt((x.value - mpc_obs.comment.x) ** 2 + (y.value - mpc_obs.comment.y) ** 2) > 1.9:
            logging.warn("Centroid shifted ({},{}) -> ({},{})".format(mpc_obs.comment.x,
                                                                      mpc_obs.comment.y,
                                                                      x.value,
                                                                      y.value))
    except Exception as ex:
        logging.error(str(ex))

    # Don't use the new X/Y for Hand measured entries.  (although call to get_observed_magnitude should have changed)
    if str(mpc_obs.note1) != "H" and not skip_centroids:
        mpc_obs.comment.x = x.value
        mpc_obs.comment.y = y.value

    try:
        mag = float(mag)
    except:
        return mpc_obs

    if math.isnan(mag):
        return mpc_obs

    if mag > 10:
        mpc_obs._band = filter_value
        mpc_obs.comment.mag = mag
        mpc_obs.comment.mag_uncertainty = merr

    # Update the mpc record magnitude if previous value existed here.
    if (mpc_obs.mag is not None or (mpc_obs.mag is None and mpc_in.comment.photometry_note[0] == "Z")) and mag > 10:
        mpc_obs.mag = mag

    return mpc_obs


def run(mpc_file, cor_file, 
        skip_discovery=True, skip_mags=False, 
        skip_centroids=False, compare_orbits=False):
    """

    :param mpc_file: A file containing the astrometric lines to be updated.
    :param cor_file: The base name for the updated astrometry and diagnostic files.
    :param skip_mags: Should we skip recomputing the magnitude of sources?
    :return: :raise ValueError: If actions on the mpc_obs indicate this is not a valid OSSOS observations
    """
    observations = mp_ephem.EphemerisReader().read(mpc_file)
    logging.debug("Read in Observations: {}".format(observations))
    original_obs = []
    modified_obs = []
    logging.info("ASTROMETRY FILE: {} --> {}.tlf".format(mpc_file, cor_file))
    for mpc_in in observations:
      try:
        if not isinstance(mpc_in.comment, mp_ephem.ephem.OSSOSComment):
            logging.info(type(mpc_in.comment))
            logging.info("Skipping: {}".format(mpc_in.to_string()))
            continue
        if ((skip_discovery and mpc_in.discovery) or
                (not skip_discovery and not mpc_in.discovery)):
            logging.info("Discovery mis-match")
            logging.info("Skipping: {}".format(mpc_in.to_string()))
            continue
        logging.info("="*220)
        logging.info("   orig: {}".format(mpc_in.to_string()))
        if mpc_in.comment.astrometric_level == 4:
            logging.info("Already at maximum AstLevel, skipping.")
            continue
        if mpc_in.null_observation:
            logging.info("Skipping NULL observation.")
            continue
        mpc_obs = remeasure(mpc_in)
        logging.info("new wcs: {}".format(mpc_obs.to_string()))

        if not skip_mags:  
            # and not mpc_obs.comment.photometry_note[0] == "Z":
            mpc_mag = remeasure(recompute_mag(mpc_obs,
                                              skip_centroids=skip_centroids),
                                reset_pixel_coordinates=not skip_centroids)
        else:
            mpc_mag = mpc_obs

        sep = mpc_in.coordinate.separation(mpc_mag.coordinate)
        if sep > TOLERANCE:
            logging.error("Large offset: {} arc-sec".format(sep))
            logging.error("orig: {}".format(mpc_in.to_string()))
            logging.error(" new: {}".format(mpc_mag.to_string()))
            new_comment = "BIG SHIFT HERE"
            mpc_mag.comment.comment = mpc_mag.comment.comment + " " + new_comment
        logging.info("new cen: {}".format(mpc_mag.to_string()))
        original_obs.append(mpc_in)
        modified_obs.append(mpc_mag)
        logging.info("="*220)
      except:
        logging.error("Skipping: {}".format(mpc_in))

    optr = open(cor_file + ".tlf", 'w')
    for idx in range(len(modified_obs)):
        inp = original_obs[idx]
        out = modified_obs[idx]
        if inp != out:
            optr.write(out.to_tnodb()+"\n")
    optr.close()

    if not compare_orbits:
        return True
    try:
       compare_orbits(original_obs, modified_obs, cor_file)
    except Exception as ex:
       logging.error("Orbit comparison failed: {}".format(ex))
    logging.info("="*220)

    return True


def compare_orbits(original_obs, modified_obs, cor_file):
    """Compare the orbit fit given the oringal and modified astrometry."""

    origin = orbfit.Orbfit(original_obs)
    modified = orbfit.Orbfit(modified_obs)

    orbpt = file(cor_file+".orb", 'w')

    # Dump summaries of the orbits
    orbpt.write("#"*80+"\n")
    orbpt.write("# ORIGINAL ORBIT\n")
    orbpt.write(origin.summarize()+"\n")
    orbpt.write("#"*80+"\n")
    orbpt.write("# MODIFIED ORBIT\n")
    orbpt.write(modified.summarize()+"\n")
    orbpt.write("#"*80+"\n")

    # Create a report of the change in orbit parameter uncertainty
    for element in ['a', 'e', 'inc', 'om', 'Node', 'T']:
        oval = getattr(origin, element).value
        doval = getattr(origin, "d"+element).value
        mval = getattr(modified, element).value
        dmval = getattr(modified, "d"+element).value
        precision = max(int(-1*math.floor(math.log10(dmval))), int(-1*math.floor(math.log10(doval)))) + 1
        precision = max(0, precision)
        vdigits = 12
        ddigits = 6
        vpadding = " "*int(vdigits-precision)
        dpadding = " "*int(ddigits-precision)

        orbpt.write("{element:>5s}: "
                    "{oval[0]:>{vdigits}.{vdigits}}.{oval[1]:<{precision}.{precision}} {vpadding} +/- "
                    "{doval[0]:>{ddigits}.{ddigits}}.{doval[1]:<{precision}.{precision}} {dpadding} ==> "
                    "{mval[0]:>{vdigits}.{vdigits}}.{mval[1]:<{precision}.{precision}} {vpadding} +/- "
                    "{dmval[0]:>{ddigits}.{ddigits}}.{dmval[1]:<{precision}.{precision}}\n".format(
            element=element,
            dpadding=dpadding,
            vpadding=vpadding,
            vdigits=vdigits,
            ddigits=ddigits,
            oval="{:12.12f}".format(oval).split("."),
            doval="{:12.12f}".format(doval).split("."),
            mval="{:12.12f}".format(mval).split("."),
            dmval="{:12.12f}".format(dmval).split("."),
            precision=precision)
        )

        delta = math.fabs(oval - mval)
        if delta > 3.5 * doval:
            logging.warn("large delta for element {}: {} --> {}".format(element, oval, mval))


    # Compute the stdev of the residuals and report the change given the new observations
    orbpt.write("*"*80+"\n")
    orbpt.write("Change in orbital parameters \n")
    sep = "Change in scatter between initial and recalibrated obseravtions. \n"
    for orb in [origin, modified]:
        orbpt.write(sep)
        sep = "\n ==> becomes ==> \n"
        residuals = orb.residuals
        dra = []
        ddec = []
        mags = {}
        for observation in orb.observations:
            if not observation.null_observation:
                dra.append(observation.ra_residual)
                ddec.append(observation.dec_residual)
                filter = observation.band
                if filter is not None:
                    if filter not in mags:
                        mags[filter] = []
                    try:
                        mags[filter].append(float(observation.mag))
                    except:
                        pass
                if observation.comment.plate_uncertainty * 5.0 < \
                        ((observation.ra_residual ** 2 + observation.dec_residual ** 2) ** 0.5):
                    logging.warn("LARGE RESIDUAL ON: {}".format(observation.to_string()))
                    logging.warn("Fit residual  unreasonably large.")
        dra = numpy.array(dra)
        ddec = numpy.array(ddec)
        merr_str = ""
        for filter in mags:
            mag = numpy.percentile(numpy.array(mags[filter]), (50))
            mags[filter] = numpy.percentile(numpy.array(mags[filter]), (5,95))
            merr = (mags[filter][1] - mags[filter][0])/6.0
            merr_str += " {}: {:8.2f} +/- {:8.2f}".format(filter, mag, merr)
        orbpt.write("ra_std:{:8.4} dec_std:{:8.4} mag: {}".format(dra.std(), ddec.std(), merr_str))
    orbpt.write("\n")
    orbpt.close()


def main():
    description = """This program takes as input a set of OSSOS measurements and adjusts the astrometric and photometric
    entries to be consistent with the current best estimate for the astrometric and photometric calibrations.
    """
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('ast_file', help="An MPC file to update.")
    parser.add_argument('--discovery', help="Only process the discovery images.", action='store_true', default=False)
    parser.add_argument('--result_base_name', help="base name for remeasurement results (defaults to basename of input)",
                        default=None)
    parser.add_argument('--skip-mags', action="store_true", help="Recompute magnitudes.", default=False)
    parser.add_argument('--skip-centroids', action="store_true", help="Recompute centroids.", default=False)
    parser.add_argument('--compare-orbits', action='store_true', help="Compute/Compare pre and post remeasure orbits?", default=False)
    parser.add_argument('--debug', action='store_true')

    args = parser.parse_args()

    level = logging.INFO
    if args.debug:
        level = logging.DEBUG

    import coloredlogs
    logger = logging.getLogger('update_astrom')
    coloredlogs.install(level=level)

    if args.result_base_name is None:
        base_name = os.path.splitext(os.path.basename(args.ast_file))[0]
    else:
        base_name = args.result_base_name
    run(args.ast_file, base_name, 
        skip_discovery=not args.discovery, 
        skip_mags=args.skip_mags, 
        skip_centroids=args.skip_centroids, 
        compare_orbits=args.compare_orbits)


if __name__ == '__main__':
    sys.exit(main())



#!python
from multiprocessing import Pool
from ossos import mop_file, storage
import os
from astropy.io import fits
import numpy
from vos import Client
import sys
from ossos import junk_keywords
from ossos import util
import argparse
import logging

client = Client()

RA = 'RA_J2000'
DEC = 'DE_J2000'
del_keyword_list = junk_keywords.bad_keywords_list
task = 'slow'
dependency = 'step1'

def run(expnum, ccd, version='p', prefix='', dry_run=False, force=False):

    message = 'success'

    if storage.get_status(task, prefix, expnum, version=version, ccd=ccd) and not force:
        logging.info("{} completed successfully for {} {} {} {}".format(task, prefix, expnum, version, ccd))
        return

    with storage.LoggingManager(task=task, prefix=prefix, expnum=expnum, ccd=ccd, version=version, dry_run=dry_run):
        try:
            if not storage.get_status(dependency, prefix, expnum, "p", ccd=ccd):
                raise IOError("{} not yet run for {}".format(dependency, expnum))


            header = storage.get_astheader(expnum, ccd)
            datasec = storage.datasec_to_list(header.get('DATASEC', '[80:2080,30,4160]'))
            try:
                fwhm = "{:5.2f}".format(storage.get_fwhm(expnum, ccd))
            except:
                fwhm = 'unknown'
            for keyword in del_keyword_list:
                try:
                    del(header[keyword])
                except:
                    pass
            header['FWHM'] = (fwhm, 'FWHM in pixels')
            header['EXTNAME'] = 'header'
            primary_hdu = fits.PrimaryHDU(header=header)
            hdu_list = fits.HDUList([primary_hdu, ])
            for ext in ['jmp', 'matt']:
                extension = 'obj.'+ext
                name = "{}p{:02d}.{}".format(expnum, ccd, extension)
                try:
                    os.unlink(name)
                    os.unlink(name+".fits")
                except:
                    pass
                logging.info("Retrieving {}".format(name))
                obj_file = mop_file.Parser(expnum, ccd, extension)
                obj_file.parse()
        
                t = numpy.all([datasec[0] < obj_file.data['X'], obj_file.data['X'] < datasec[1],
                               datasec[2] < obj_file.data['Y'], obj_file.data['Y'] < datasec[3]], axis=0)
                logging.info("Source remaining after datasec cut: {} of {}".format(len(obj_file.data[t]['X']), len(t)))
                table_hdu = fits.table_to_hdu(obj_file.data[t])
                table_hdu.header['CATALOG'] = name
                table_hdu.header['EXTNAME'] = ext
                hdu_list.append(table_hdu)
                del table_hdu
                del obj_file
                os.unlink(name)

            name = "{}p{:02d}.{}".format(expnum, ccd, 'obj.fits')
            if os.access(name, os.F_OK):
                os.unlink(name)
            hdu_list.writeto(name)
            uri = storage.dbimages_uri(expnum, ccd, 'p', ext=".obj.fits")
            logging.info(name+" -> "+uri)
            count = 0
            with open(name):
              while True:
                count += 1
                logging.info("Copy attempt {}".format(count))
                try:
                    storage.copy(name, uri)
                    os.unlink(name)
                    break
                except Exception as ex:
                    if count > 10:
                        raise ex

            logging.info(message)
        except Exception as e:
            message = str(e)
            logging.error(message)
        
        if not dry_run:
            storage.set_status(task, prefix, expnum, version=version, ccd=ccd, status=message)


if __name__ == '__main__':


    parser = argparse.ArgumentParser(
        description='Run FITS catalog builder chunk of the OSSOS pipeline')

    parser.add_argument('--ccd', '-c',
                        action='store',
                        type=int,
                        dest='ccd',
                        default=None,
                        help='which ccd to process, default is all')
    parser.add_argument("--dbimages",
                        action="store",
                        default="vos:OSSOS/dbimages",
                        help='vospace dbimages containerNode')
    parser.add_argument("expnum",
                        type=int,
                        nargs='+',
                        help="expnum(s) to process")
    parser.add_argument("--dry-run",
                        action="store_true",
                        help="DRY RUN, don't copy results to VOSpace, implies --force")

    parser.add_argument("--fk", action="store_true", help="Run fk images")

    parser.add_argument("--type", "-t", choices=['o', 'p', 's'],
                        help="which type of image: o-RAW, p-ELIXIR, s-SCRAMBLE", default='p')
    parser.add_argument("--verbose", "-v",
                        action="store_true")
    parser.add_argument("--force", default=False,
                        action="store_true")
    parser.add_argument("--debug", "-d",
                        action="store_true")

    cmd_line = " ".join(sys.argv)
    args = parser.parse_args()

    util.set_logger(args)
    logging.info("Started {}".format(cmd_line))

    prefix = (args.fk and 'fk') or ''

    storage.DBIMAGES = args.dbimages

    for expnum in args.expnum:
        if args.ccd is None:
            ccdlist = storage.get_ccdlist(expnum)
        else:
           ccdlist = [args.ccd]

        for ccd in ccdlist:
            run(expnum, ccd)



#!python

import argparse
import logging
import os
import re
import sys

from ossos import mpc
from astropy import units
from astropy.time import Time
from astropy.coordinates.angles import Angle
from ossos.gui import config

rename_map = {}


def load_observations(xxx_todo_changeme, path, filenames):
    """
    Returns a provisional name based dictionary of observations of the object.
    Each observations is keyed on the date. ie. a dictionary of dictionaries.

    @param path: the directory where filenames are.
    @type path str
    @param filenames: list of files in path.
    @type filenames list
    @rtype  None
    """
    (observations, regex, rename) = xxx_todo_changeme
    for filename in filenames:
        if re.search(regex, filename) is None:
            logging.error("Skipping {}".format(filename))
            continue
        obs = mpc.MPCReader().read(os.path.join(path, filename))
        for ob in obs:
            if "568" not in ob.observatory_code:
                continue
            if not isinstance(ob.comment, mpc.OSSOSComment):
                continue
            if ob.date < Time("2013-01-01 00:00:00"):
                continue
            if rename:
                new_provisional_name = os.path.basename(filename)
                new_provisional_name = new_provisional_name[0:new_provisional_name.find(".")]
                rename_map[ob.provisional_name] = new_provisional_name
            try:
                key1 = ob.comment.frame.split('p')[0]
            except Exception as ex:
                logger.warning(str(ex))
                logger.warning(ob.to_string())
                continue
            key2 = ob.provisional_name
            if key1 not in observations:
                observations[key1] = {}
            if key2 in observations[key1]:
                if observations[key1][key2]:
                    continue
                if not observation.null_observation:
                    logger.error(filename)
                    logger.error(str(observations[key1][key2]))
                    raise ValueError("conflicting observations for {} in {}".format(key2, key1))
            observations[key1][key2] = ob

def main():

    description = ("Given two directories containing astrometric observations build"
                   "file that reports the newly measured astrometry.")

    epilog = """This script is intended to provide a simple way of gathering a list of new astrometry for reporting.
Astrometry is reported to the a central service for distribution (dbase fro the OSSOS project, or the
Minor Planet Center, for example). These central system do not want previously reported astrometry in
new submission files. Also, when a line of astrometry needs to be changed there is, nominally, a separate channel
for this reporting.  This script attempts to provide an automated approach to creating these reports.

In addition to the reported astrometry, each file should also contain some comments on the nature of the observations
being submitted. This script, being part of the OSSOS pipeline, provides these comments in a way that is consistent
with the expectations of the OSSOS reporting structure.  The script should build these comments in a reasonably
auto-magical way.
"""

    app_config = config.AppConfig()
    parser = argparse.ArgumentParser(description=description, epilog=epilog)
    parser.add_argument("existing_astrometry_directory",
                        help=("Directory containing previously reported astrometry."
                              "Files can be in dbase ast, validate ast or OSSOS mpc format"))
    parser.add_argument("--idx_filename",
                        default=None,
                        help="File that has the MOP/OSSOS name mapping index.")
    parser.add_argument("new_astrometry_directory",
                        help=("Directory containing astrometry to be searched for new lines to report."
                              "Files should be in validate ast format."
                              "Only observations from a single observatory should be present in these files."))
    parser.add_argument("report_file",
                        help="Name of file that new lines will be reported to")
    parser.add_argument("--rename", action="store_true",
                        help="Rename objects in new measurement files based on the name of the file")
    parser.add_argument("--new_name_regex",
                        default='.*\.ast',
                        help="Only load new observations where provisional name matches")
    parser.add_argument("--existing_name_regex",
                        default='.*\.ast',
                        help="Only load existing observations where provisional name matches")
    parser.add_argument("--start_date", type=mpc.get_date,
                        help="Include observations taken on or after this date in report.")
    parser.add_argument("--end_date", type=mpc.get_date,
                        help="Include observation taken on or before this date in report.")
    parser.add_argument("--replacement", action='store_true', help="select replacement lines")
    parser.add_argument("--tolerance", default=1,
                        help="""tolerance is a flag, if 1 then only report lines that have different
                        positions and/flux, if -1 then any change cuases a report line""")
    parser.add_argument("--COD", default=None,
                        help="Observatory code for report file.")
    parser.add_argument("-q",
                        action="store_true", help="Run quiet")
    parser.add_argument("--OBS", 
                        action="append",
                        help="Names of observers, multiple allowed")
    parser.add_argument("--MEA", 
                        action="append",
                        help="Names of measures, multiple allowed")
    parser.add_argument("--TEL", default=app_config.read("TELESCOPE"),
                        action="store")
    parser.add_argument("--NET", default=app_config.read("ASTROMETRIC_NETWORK"),
                        action="store")


    args = parser.parse_args()
    tolerance = Angle(float(args.tolerance) * units.arcsec)

    logger = logging.getLogger('reporter')
    if args.q:
        logger.setLevel(logging.CRITICAL)

    if args.idx_filename is None:
        args.idx_filename = os.path.dirname(args.existing_astrometry_directory) + "/idx/file.idx"

    idx = mpc.Index(args.idx_filename)

    existing_observations = {}
    os.path.walk(args.existing_astrometry_directory, load_observations,
                 (existing_observations, args.existing_name_regex, False))
    logger.info("Loaded existing observations for {} objects.\n".format(len(existing_observations)))

    new_observations = {}
    os.path.walk(args.new_astrometry_directory, load_observations,
                 (new_observations, args.new_name_regex, args.rename))
    logger.info("Loaded new observations for {} objects.\n".format(len(new_observations)))

    report_observations = {}
    for frame1 in new_observations:
        for name1 in new_observations[frame1]:
            observation1 = new_observations[frame1][name1]
            logger.warning("Checking {}".format(observation1.to_string()))
            assert isinstance(observation1, mpc.Observation)
            if ((args.start_date is None or args.start_date.jd < observation1.date.jd)
                    and (args.end_date is None or args.end_date.jd > observation1.date.jd)):
                report = True
                replacement = False
                if frame1 in existing_observations:
                    for name2 in existing_observations[frame1]:
                        observation2 = existing_observations[frame1][name2]
                        assert isinstance(observation2, mpc.Observation)
                        replacement = False
                        if idx.is_same(observation2.provisional_name, observation1.provisional_name):
                            if ((tolerance < 0 and observation1 != observation2) or
                                    (observation1.mag != observation2.mag or
                                        observation1.ra != observation2.ra or
                                        observation1.dec != observation2.dec)):
                                if not observation2.discovery:
                                    replacement = True
                                else:
                                    print("discovery")
                            else:
                                report = False
                            break
                if report and replacement == args.replacement:
                    logger.warning("Adding {} on {} to report".format(name1, observation1.date))
                    report_observations[name1] = report_observations.get(name1, [])
                    report_observations[name1].append(observation1)

    if not len(report_observations) > 0:
        logger.warning("No observations matched criterion.")
        sys.exit(0)

    if args.report_file == "-":
        outfile = sys.stdout
    else:
        outfile = open(args.report_file, 'w')

    full_observation_list = []
    for name in report_observations:
        full_observation_list.extend(report_observations[name])

    observers = args.OBS is not None and len(args.OBS) > 0 and args.OBS or None
    measurers = args.MEA is not None and len(args.MEA) > 0 and args.MEA or None
    network = len(args.NET) > 0 and args.NET or None
    telescope = len(args.TEL) > 0 and args.TEL or None

    outfile.write(mpc.make_tnodb_header(full_observation_list, 
                                        observers=observers,
                                        measurers=measurers,
                                        telescope=telescope,
                                        astrometric_network=network))
    outfile.write("\n")

    for name in report_observations:
        # sorted("This is a test string from Andrew".split(), key=str.lower)
        report_observations[name].sort(key=lambda x: x.date.jd)
        for observation in report_observations[name]:
            if name in rename_map:
                observation.provisional_name = rename_map[name]
            outfile.write(observation.to_tnodb() + "\n")
        outfile.write("\n")
    outfile.close()


if __name__ == '__main__':
    sys.exit(main())



###############################################################################
#                                                                            ##
# Copyright 2013 by its authors                                              ##
# See COPYING, AUTHORS                                                       ##
#                                                                            ##
# This file is part of OSSOS Moving Object Pipeline (OSSOS-MOP)              ##
#                                                                            ##
#    OSSOS-MOP is free software: you can redistribute it and/or modify       ##
#    it under the terms of the GNU General Public License as published by    ##
#    the Free Software Foundation, either version 3 of the License, or       ##
#    (at your option) any later version.                                     ##
#                                                                            ##
#    OSSOS-MOP is distributed in the hope that it will be useful,            ##
#    but WITHOUT ANY WARRANTY; without even the implied warranty of          ##
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           ##
#    GNU General Public License for more details.                            ##
#                                                                            ##
#    You should have received a copy of the GNU General Public License       ##
#    along with OSSOS-MOP.  If not, see <http://www.gnu.org/licenses/>.      ##
#                                                                            ##
###############################################################################
"""run step3 of the OSSOS pipeline."""

import os
import argparse
import logging
import sys
from ossos import util
from ossos import storage

_RATE_MIN = 0.5
_RATE_MAX = 15.0
_ANGLE_CENTRE = 23.0
_ANGLE_WIDTH = 30.0

task = util.task()
dependency = 'step2'


def run(expnums, ccd, version, rate_min,
        rate_max, angle, width, field=None, prefix=None, dry_run=False, force=False):
    """run the actual step2  on the given exp/ccd combo"""

    jmp_args = ['step3jmp']
    matt_args = ['step3matt']

    if storage.get_status(task, prefix, expnums[0], version=version, ccd=ccd) and not force:
        logging.info("{} completed successfully for {}{}{}{:02d}".format(
            task, prefix, expnums[0], version, ccd))
        return

    with storage.LoggingManager(task, prefix, expnums[0], ccd, version, dry_run):
        try:
            if not storage.get_status(dependency, prefix, expnums[0], version=version, ccd=ccd):
                raise IOError(35, "Cannot start {} as {} not yet completed {}{}{}{:02d}".format(
                    task, dependency, prefix, expnums[0], version, ccd))
            # Default message is success, message gets overwritten with failure messages.
            message = storage.SUCCESS

            idx = 0
            cmd_args = []
            for expnum in expnums:
                idx += 1
                for ext in ['unid.jmp', 'unid.matt',
                            'trans.jmp']:
                    storage.get_file(expnum, ccd=ccd, version=version, ext=ext, prefix=prefix)
                image = os.path.splitext(os.path.basename(storage.get_uri(expnum,
                                                                          ccd,
                                                                          version=version,
                                                                          prefix=prefix)))[0]
                cmd_args.append('-f%d' % idx)
                cmd_args.append(image)

            cmd_args.extend(['-rn', str(rate_min),
                             '-rx', str(rate_max),
                             '-a', str(angle),
                             '-w', str(width)])
            jmp_args.extend(cmd_args)
            matt_args.extend(cmd_args)
            logging.info(util.exec_prog(jmp_args))
            logging.info(util.exec_prog(matt_args))

            if dry_run:
                return

            if field is None:
                field = str(expnums[0])

            # Make sure a dbimages destination exists for this file.
            storage.mkdir(os.path.dirname(storage.get_uri(field,
                                                          ccd=ccd,
                                                          version=version,
                                                          prefix=prefix)))

            for ext in ['moving.jmp', 'moving.matt']:
                uri = storage.get_uri(field,
                                      ccd=ccd,
                                      version=version,
                                      ext=ext,
                                      prefix=prefix)
                filename = '%s%d%s%s.%s' % (prefix, expnums[0],
                                            version,
                                            str(ccd).zfill(2),
                                            ext)
                storage.copy(filename, uri)

        except Exception as ex:
            message = str(ex)
            logging.error(message)

        storage.set_status(task, prefix, expnums[0], version=version, ccd=ccd, status=message)

    return


def main():
    # Must be running as a script

    parser = argparse.ArgumentParser(
        description='Run step3jmp and step3matt on a given triple.')

    parser.add_argument("--ccd", "-c",
                        action="store",
                        default=None,
                        type=int,
                        dest="ccd")
    parser.add_argument("--dbimages",
                        action="store",
                        default="vos:OSSOS/dbimages",
                        help='vospace dbimages containerNode')
    parser.add_argument("expnums",
                        type=int,
                        nargs=3,
                        help="3 expnums to process")
    parser.add_argument("--version",
                        action='version',
                        version='%(prog)s 1.0')
    parser.add_argument('-t', '--type',
                        help='which type of image to process',
                        choices=['s', 'p', 'o'],
                        default='p')
    parser.add_argument('--fk', help='Do fakes?', default=False, action='store_true')
    parser.add_argument('--field',
                        help='a string that identifies which field is being searched',
                        nargs='?')
    parser.add_argument('--no-sort',
                        help='preserve input exposure order',
                        action='store_true')
    parser.add_argument("--verbose", "-v",
                        action="store_true")
    parser.add_argument("--debug", action="store_true")
    parser.add_argument("--rate_min", default=_RATE_MIN,
                        help='minimum rate to accept',
                        type=float)
    parser.add_argument('--rate_max', default=_RATE_MAX,
                        help='maximum rate to accept',
                        type=float)
    parser.add_argument('--angle', default=_ANGLE_CENTRE,
                        help='angle of x/y motion, West is 0, North 90',
                        type=float)
    parser.add_argument('--width', default=_ANGLE_WIDTH,
                        help='openning angle of search cone',
                        type=float)
    parser.add_argument("--dry-run", action="store_true", help="do not copy to VOSpace, implies --force")
    parser.add_argument("--force", action="store_true")

    cmd_line = " ".join(sys.argv)
    args = parser.parse_args()
    util.set_logger(args)
    logging.info("Starting {}".format(cmd_line))

    storage.DBIMAGES = args.dbimages
    prefix = (args.fk and 'fk') or ''
    expnums = args.expnums
    version = args.type

    if args.ccd is None:
        expnum = min(expnums)
        if int(expnum) < 1785619:
            # Last exposures with 36 CCD Megaprime
            ccdlist = list(range(0, 36))
        else:
            # First exposrues with 40 CCD Megaprime
            ccdlist = list(range(0, 40))
    else:
        ccdlist = [args.ccd]

    if not args.no_sort:
        args.expnums.sort()

    for ccd in ccdlist:
        run(expnums, ccd, version=version,
            rate_min=args.rate_min,
            rate_max=args.rate_max,
            angle=args.angle,
            width=args.width,
            field=args.field,
            prefix=prefix,
            force=args.force,
            dry_run=args.dry_run)


if __name__ == '__main__':
    sys.exit(main())





#!python
"""Retrieval of cutouts of the FITS images associated with the OSSOS detections.

Takes a directory of .ast file (in dbase format) as input

An Example URL for cutouts
http://www.canfar.phys.uvic.ca/vospace/auth/synctrans?TARGET=vos://cadc.nrc.ca~
vospace/OSSOS/dbimages/1625356/1625356p.fits&DIRECTION=pullFromVoSpace&PROTOCOL=
ivo://ivoa.net/vospace/core%23httpget&view=cutout&cutout=CIRCLE+ICRS+242.1318+-1
2.4747+0.05
"""

import argparse
import logging
import os
import sys

from astropy import units
from astropy.units import Quantity
from astropy.io import fits

from ossos import (mpc, storage, parameters)
storage.FITS_EXT = ".fits"

def cutout(obj, obj_dir, radius):

    cutout_listing = storage.listdir(obj_dir, force=True)
    for obs in obj.mpc_observations:
        if obs.null_observation:
            logging.debug('skipping: {}'.format(obs))
            continue
        if obs.date > parameters.SURVEY_START:  # can't make postage stamps of earlier linkages
            # can't parse for an obs.comment's exposure number if no obs.comment exists
            try:
                parts = storage.frame2expnum(obs.comment.frame)
            except Exception as ex:
                logging.error("Skipping: {}\nFailed to map comment.frame to expnum: {}".format(obs, ex))
                continue
            uri = storage.get_uri(parts['expnum'], version=parts['version'])
            sky_coord = obs.coordinate  # Using the WCS rather than the X/Y (X/Y can be unreliable over the whole survey)
            postage_stamp_filename = "{}_{:11.5f}_{:09.5f}_{:+09.5f}.fits".format(obj.provisional_name,
                                                                                  obs.date.mjd,
                                                                                  obs.coordinate.ra.degree,
                                                                                  obs.coordinate.dec.degree)

            if postage_stamp_filename in cutout_listing:
               # skipping existing cutouts
                continue 

            # ast_header = storage._get_sghead(parts['expnum'])
            while True:
              try:
                hdulist = storage.ra_dec_cutout(uri, sky_coord, radius, update_wcs=True)

                with open(postage_stamp_filename, 'w') as tmp_file:
                    hdulist.writeto(tmp_file, overwrite=True, output_verify='fix+ignore')
                    storage.copy(postage_stamp_filename, obj_dir + "/" + postage_stamp_filename)
                os.unlink(postage_stamp_filename)  # easier not to have them hanging around
              except OSError as e:  # occasionally the node is not found: report and move on for later cleanup
                logging.error("OSError: ->"+str(e))
              except Exception as e:
                logging.error("Exception: ->"+str(e))
                continue
              break


def main():
    parser = argparse.ArgumentParser(
        description='Parse a directory of TNO .ast files and create links in the postage stamp directory '
                    'that allow retrieval of cutouts of the FITS images associated with the OSSOS detections. '
                    'Cutouts are defined on the WCS RA/DEC of the object position.')

    parser.add_argument("version",
                        help="The OSSOS data release version these stamps should be assigned to. e.g. v8")
    parser.add_argument("astdir", help="Directory containing the input .ast files", action="store", default="ast/")
    parser.add_argument("--ossin",
                        action="store",
                        default="vos:OSSOS/0_OSSOSreleases/OSSOS",
                        help="The vospace containerNode that clones ossin dbaseclone"
                             "holding the .ast files of astrometry/photometry measurements.")
    parser.add_argument("--blocks", "-b",
                        action="store",
                        default=['o3e', 'o3o', 'o3l', 'o4h', 'o5p', 'o5m', 'o5s', 'o5t', 'o5c', 'o5d'],
                        choices=["o3e", "o3o", "Col3N", 'o3l', 'o4h', 'o5m', 'o5p', 'o5s', 'o5t', 'o5c', 'o5d'],
                        help="Prefixes of object designations to include.")
    parser.add_argument("--radius", '-r',
                        # FIXME: figure out how to assign this a unit.degree before storage
                        action='store',
                        default=36 * units.arcsec,  # about 200 px square
                        help='Radius (arcsec) of circle of cutout postage stamp.')
    parser.add_argument("--debug", "-d",
                        action="store_true")
    parser.add_argument("--verbose", "-v",
                        action="store_true")
    parser.add_argument("--recheck",
                        default=None,
                        action="store",
                        help="A tuple of TNO IDs to rerun")


    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    elif args.verbose:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.ERROR)


    astdir = args.astdir
    flist = os.listdir(astdir)
    if args.recheck:
        flist = [args.recheck + '.ast']

    for fn in flist:
        if not fn.endswith('.ast'):
            continue
        for block in args.blocks:
            if fn.startswith(block):
                obj_dir = '{}/{}/{}'.format(storage.POSTAGE_STAMPS, args.version, fn.partition('.')[0]) # obj.provisional_name
                logging.info("Processing astrometric files in {}".format(obj_dir))
                storage.mkdir(obj_dir)
                obj = mpc.MPCReader(astdir + fn)
                # assert storage.exists(obj_dir, force=True)
                sys.stderr.write('{} beginning...'.format(obj.provisional_name))
                # if int(obj.provisional_name[3:]) == 49:
                assert isinstance(args.radius, Quantity)
                cutout(obj, obj_dir, args.radius)
                sys.stderr.write('{} complete.\n\n'.format(obj.provisional_name))


if __name__ == '__main__':
    main()



import json
import argparse
import sys
from astropy.coordinates import SkyCoord


class Program(object):
    def __init__(self, runid="17AF22", pi_login="marsset"):
        self.config = {"runid": runid,
                       "pi_login": pi_login,
                       "program_configuration": {"targets": [],
                                                 "observing_blocks": [],
                                                 "observing_groups": []
                                                 }}

    def add_target(self, target):
        self.config["program_configuration"]["targets"].append(target)

    def add_observing_block(self, observing_block):
        self.config["program_configuration"]["observing_blocks"].append(observing_block)

    def add_observing_group(self, observing_group):
        self.config["program_configuration"]["observing_groups"].append(observing_group)


class Target(object):
    def __init__(self, filename=None):
        self.config = json.load(open(filename))

    @property
    def token(self):
        return self.config["identifier"]["client_token"]

    @property
    def mag(self):
        return self.config["moving_target"]["ephemeris_points"][0]["mag"]

    @property
    def coordinate(self):
        return SkyCoord(self.config["moving_target"]["ephemeris_points"][0]["coordinate"]["ra"],
                        self.config["moving_target"]["ephemeris_points"][0]["coordinate"]["dec"],
                        unit='degree')


class ObservingBlock(object):
    def __init__(self, client_token, target_token):
        self.config = {"identifier": {"client_token": client_token},
                       "target_identifier": {"client_token": target_token},
                       "constraint_identifiers": [{"server_token": "C1"}],
                       "instrument_config_identifiers": [{"server_token": "I1"}]}

    @property
    def token(self):
        return self.config["identifier"]["client_token"]


class ObservingGroup(object):
    def __init__(self, client_token):
        self.config = {"identifier": {"client_token": client_token},
                       "observing_block_identifiers": []}

    def add_ob(self, client_token):
        self.config["observing_block_identifiers"].append({"client_token": client_token})


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('pi_login')
    parser.add_argument('qrunid')
    parser.add_argument('runid')
    parser.add_argument('targets', nargs='+')
    args = parser.parse_args()

    # Break the targets into OBs based on their max mag of source in pointing.
    OB = {"I1": "r", "I2": "u"}
    OG = ["I1"] + ["I2"]*14 + ["I1"]
    
    program = Program()
    ob_tokens = {}
    runid = "17AF22"
    targets = {}
    og_idx = 0
    for filename in args.targets:
        target = Target(filename)
        targets[target.token] = target
        program.add_target(target.config)
        for idx in ["I1", "I2"]:
           ob_token = "OB-{}-{}-{}".format(runid, target.token, OB[idx])
           ob = ObservingBlock(ob_token, target.token)
           ob.config["instrument_config_identifiers"] = [{"server_token": "{}".format(idx)}]
           program.add_observing_block(ob.config)
        og_idx += 1
        og_token = "OG_{}_{}_{}_{}".format(args.runid, args.qrunid, og_idx, 0)
        sys.stdout.write("{}: ".format(og_token))
        og = ObservingGroup(og_token)
        for IC in OG:
            ob_token = "OB-{}-{}-{}".format(runid, target.token, OB[IC])
            og.add_ob(ob_token)
        program.add_observing_group(og.config)

    json.dump(program.config, open('program.json', 'w'), indent=4, sort_keys=True)



"""CFHT PH2 objects"""
import json


class PH2(object):

    def to_json_string(self):
        return json.dumps(self.to_dict())

    def from_json_string(self, value):
        parts = json.loads(value)
        for part in parts:


    def to_dict(self):
        _dict = self.__dict__
        for part in _dict:
            if isinstance(_dict[part], PH2):
                _dict[part] = _dict[part].to_dict()
            if isinstance(_dict[part], list):
                _dict[part] = [item.to_dict() for item in _dict[part]]
        return _dict


class Program(PH2):

    def __init__(self, runid, pi_login):
        self.runid = runid
        self.pi_login = pi_login
        self.program_configuration = None


class ProgramConfiguration(PH2):

    def __init__(self):
        self.targets = []
        self.observing_blocks = []
        self.observing_groups = []


class Identifier(PH2):

    def __init__(self, server_token=None, client_token=None):
        if server_token is not None:
            self.server_token = server_token
        if client_token is not None:
            self.client_token = client_token


class Coordinate(PH2):

    def __init__(self, ra, dec):
        assert isinstance(ra, float)
        assert isinstance(dec, float)
        self.ra = ra
        self.dec = dec


class EphmerisPoint(PH2):

    def __init__(self, epoch, coordinate):
        assert isinstance(epoch, float)
        self.epoch = epoch
        assert isinstance(coordinate, Coordinate)
        self.coordinate = coordinate


class Target(PH2):

    def __init__(self, identifier=None, name=None, moving_target=None, fixed_target=None):
        self.identifier = identifier
        self.name = name
        if moving_target is not None:
            assert isinstance(moving_target, MovingTarget)
            self.moving_target = moving_target
        else:
            self.fixed_target = fixed_target


class MovingTarget(PH2):

    def __init__(self, ephemeris_points=None):

        if ephemeris_points is None:
            self.ephmeris_points = []
        else:
            assert isinstance(ephemeris_points, list)
            self.ephmeris_points = ephemeris_points


class ObservingBlock(PH2):

    def __init__(self, identifier, target, constraint, instrument_config):
        """
        An Observing Block holds a chunk of information that describes an single observations of a target.

        @type identifier: Identifier
        @type target: Target
        @type constraint: Constraint
        @type instrument_config: InstrumentConfig
        @param identifier: identifier of this observing block, unique within the runid
        @param target: the target of this observing block
        @param constraint: the observing constraint to apply
        @param instrument_config: the instrument configuration to use
        """
        self.identifier = identifier
        self.target_identifier = target.identifier
        self.constraint_identifier = constraint.identifier
        self.instrument_config_identifier = instrument_config.identifier


class ObservingGroup(PH2):

    def __init__(self, identifier):
        """
        A Group of Observing Blocks that should be observed together in a sequence.

        @type identifier: Identifier
        @param identifier: A unique string within this program that labels this Observing Group
        """
        self.identifier = identifier
        self.observing_block_identifiers = []


class Constraint(PH2):

    def __init__(self, identifier):
        """
        The constaraint on the observations (IQ, Photometric, Airmass, etc). Currently only set via web form.

        @type identifier; Identifier
        @param identifier: A unique, within this program, identifier of a constraint set.
        """

        self.identifier = identifier


class InstrumentConfig(PH2):
    """
    The instrument configuration on the observations (Exposure Time, Filter, Dither Pattern)

    @type identifier; Identifier
    @param identifier: A unique, within this program, identifier of a instrument configuration set.
    """

    def __init__(self, identifier):
        self.identifier = identifier





#!python
"""
Query the CADC for all exposures that brett is the PI of and taken in 2017
observations of the TNOs listed on the command line
"""

from astropy import units
from astropy.coordinates import SkyCoord
from astropy.time import TimeDelta, Time
import numpy as np
from ossos.cameras import Camera
from ossos import mpc
import mp_ephem
from ossos.ephem_target import EphemTarget
import sys, os
import argparse
import logging
import math
from copy import deepcopy
import ephem
from astropy.table import Table
import requests
from io import StringIO

cfht = ephem.Observer()
cfht.lat = 0.344
cfht.lon = -2.707
cfht.elevation = 4100
cfht.date = '2018/03/28 20:00:00'
sun = ephem.Sun()
fb = ephem.FixedBody()


def query():
    url = """http://www.cadc-ccda.hia-iha.nrc-cnrc.gc.ca/tap/sync?LANG=ADQL&REQUEST=doQuery&QUERY=SELECT%20Observation.observationURI%20AS%20%22Preview%22%2C%20Observation.collection%20AS%20%22Collection%22%2C%20Observation.sequenceNumber%20AS%20%22Sequence%20Number%22%2C%20Plane.productID%20AS%20%22Product%20ID%22%2C%20COORD1(CENTROID(Plane.position_bounds))%20AS%20%22RA%20(J2000.0)%22%2C%20COORD2(CENTROID(Plane.position_bounds))%20AS%20%22Dec.%20(J2000.0)%22%2C%20Observation.target_name%20AS%20%22Target%20Name%22%2C%20Plane.time_bounds_lower%20AS%20%22Start%20Date%22%2C%20Plane.time_exposure%20AS%20%22Int.%20Time%22%2C%20Observation.instrument_name%20AS%20%22Instrument%22%2C%20Plane.energy_bandpassName%20AS%20%22Filter%22%2C%20Plane.calibrationLevel%20AS%20%22Cal.%20Lev.%22%2C%20Observation.type%20AS%20%22Obs.%20Type%22%2C%20Observation.proposal_id%20AS%20%22Proposal%20ID%22%2C%20Observation.proposal_pi%20AS%20%22P.I.%20Name%22%2C%20Plane.dataRelease%20AS%20%22Data%20Release%22%2C%20Observation.observationID%20AS%20%22Obs.%20ID%22%2C%20Plane.energy_bounds_lower%20AS%20%22Min.%20Wavelength%22%2C%20Plane.energy_bounds_upper%20AS%20%22Max.%20Wavelength%22%2C%20AREA(Plane.position_bounds)%20AS%20%22Field%20of%20View%22%2C%20Plane.position_bounds%20AS%20%22Polygon%22%2C%20Plane.position_sampleSize%20AS%20%22Pixel%20Scale%22%2C%20Plane.energy_resolvingPower%20AS%20%22Resolving%20Power%22%2C%20Plane.time_bounds_upper%20AS%20%22End%20Date%22%2C%20Plane.dataProductType%20AS%20%22Data%20Type%22%2C%20Observation.target_moving%20AS%20%22Moving%20Target%22%2C%20Plane.provenance_name%20AS%20%22Provenance%20Name%22%2C%20Plane.provenance_keywords%20AS%20%22Provenance%20Keywords%22%2C%20Observation.intent%20AS%20%22Intent%22%2C%20Observation.target_type%20AS%20%22Target%20Type%22%2C%20Observation.target_standard%20AS%20%22Target%20Standard%22%2C%20Plane.metaRelease%20AS%20%22Meta%20Release%22%2C%20Observation.algorithm_name%20AS%20%22Algorithm%20Name%22%2C%20Observation.proposal_title%20AS%20%22Proposal%20Title%22%2C%20Observation.proposal_keywords%20AS%20%22Proposal%20Keywords%22%2C%20Plane.position_resolution%20AS%20%22IQ%22%2C%20Observation.instrument_keywords%20AS%20%22Instrument%20Keywords%22%2C%20Plane.energy_transition_species%20AS%20%22Molecule%22%2C%20Plane.energy_transition_transition%20AS%20%22Transition%22%2C%20Observation.proposal_project%20AS%20%22Proposal%20Project%22%2C%20Plane.energy_emBand%20AS%20%22Band%22%2C%20Plane.provenance_reference%20AS%20%22Prov.%20Reference%22%2C%20Plane.provenance_version%20AS%20%22Prov.%20Version%22%2C%20Plane.provenance_project%20AS%20%22Prov.%20Project%22%2C%20Plane.provenance_producer%20AS%20%22Prov.%20Producer%22%2C%20Plane.provenance_runID%20AS%20%22Prov.%20Run%20ID%22%2C%20Plane.provenance_lastExecuted%20AS%20%22Prov.%20Last%20Executed%22%2C%20Plane.provenance_inputs%20AS%20%22Prov.%20Inputs%22%2C%20Plane.energy_restwav%20AS%20%22Rest-frame%20Energy%22%2C%20Observation.requirements_flag%20AS%20%22Quality%22%2C%20Plane.planeID%20AS%20%22planeID%22%2C%20isDownloadable(Plane.planeURI)%20AS%20%22DOWNLOADABLE%22%2C%20Plane.planeURI%20AS%20%22CAOM%20Plane%20URI%22%20FROM%20caom2.Plane%20AS%20Plane%20JOIN%20caom2.Observation%20AS%20Observation%20ON%20Plane.obsID%20%3D%20Observation.obsID%20WHERE%20%20(%20Plane.calibrationLevel%20%3D%20'1'%20AND%20Observation.instrument_name%20%3D%20'MegaPrime'%20AND%20Observation.collection%20%3D%20'CFHT'%20AND%20INTERSECTS(%20INTERVAL(%2057785.0%2C%201.7976931348623157E308%20)%2C%20Plane.time_bounds%20)%20%3D%201%20AND%20lower(Observation.proposal_pi)%20LIKE%20'%25gladman%25'%20AND%20%20(%20Plane.quality_flag%20IS%20NULL%20OR%20Plane.quality_flag%20!%3D%20'junk'%20)%20)&FORMAT=votable"""
    return Table.read(StringIO(requests.get(url).content))


def coverage(orbits, locations):
    """

    @param orbits: a dictionary of orbits to optimize the point of.
    @param locations: list of CFHT pointings.
    @return:
    """

    # For each required target find the pointing that will include the largest number of other required targets
    # and then tweak that specific pointing to include the maximum number of secondary targets.
    for location in locations:
        p = Camera(SkyCoord(location["RA__J2000.0_"], location["Dec.__J2000.0_"], unit=('degree', 'degree')))
        for kbo_name in orbits:
            orbits[kbo_name].predict(Time(location["Start_Date"], format='mjd'))
            if p.polygon.isInside(orbits[kbo_name].coordinate.ra.degree,
                                  orbits[kbo_name].coordinate.dec.degree):
                print((location['Target_Name'], kbo_name))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('pointing_objects',
                        type=str,
                        help="Name of file that contains list of objects to try and observe.")
    parser.add_argument('--runid',
                        help="CFHT ID for this QUEUE Run",
                        default='16BQ17')

    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO)

    logging.info("Optimized to include as many objects in {} as possible".format(args.pointing_objects))

    filenames = open(args.pointing_objects).readlines()

    orbits = {}
    tokens = []

    # load the orbits of all objects of interest.
    for filename in filenames:
        filename = filename.strip()
        token = os.path.splitext(os.path.basename(filename))[0]
        abg_filename = os.path.splitext(filename)[0]+".abg"
        orbits[token] = mp_ephem.BKOrbit(None, ast_filename=filename, abg_file=abg_filename)
        tokens.append(token)

    coverage(orbits, query())


if __name__ == '__main__':
    sys.exit(main())






__author__ = 'Michele Bannister   git:@mtbannister'

import sys
from ossos import storage

field = sys.argv[1]  # format as e.g. AE, AO, BL

outfile = 'planning/15B{}_triplets_details.txt'.format(field)

with open('planning/triplet_and_processing_notes/{}_15B_discovery_expnums.txt'.format(field), 'r') as infile:
    # it = ims.images

    with open(outfile, 'w') as ofile:
        ofile.write(
            'Expnum RA DEC MJD_middle Exptime_sec plant_xmin_px plant_xmax_px plant_ymin_px plant_ymax_px '
            'search_xmin_px search_xmax_px search_ymin_px search_ymax_px\n'.format())

    for triplet in infile.readlines():
        with open(outfile, 'a') as ofile:  # blank line between triplets
            ofile.write('{}'.format(triplet.split()[3]))

        for expnum in triplet.split()[0:3]:
            # expnum, ra, dec, obs_end, mjd_middle, exptime (seconds)
            header = storage.get_header(storage.get_uri(expnum))
            # print header
            # JM wants MJD_middle only: calculate it from the midpoint between the start and end
            mjd_mid = header['MJDATE'] + (header['MJDEND'] - header['MJDATE']) / 2.
            retval = [expnum, header['CRVAL1'], header['CRVAL2'], mjd_mid, header['EXPTIME']]

            # now the area in which we planted TNOs for characterisation (hardwired in plant.csh's calls to kbo_gen)
            # chosen to match that in header keyword DATASEC = [33:2080,1:4612] / Imaging area of the entire CCD in
            # raw frame
            if field == 'E' or 'O':
                xmin = 66  # ya, we don't know why it got hardwired this way initially. 1.6% underplanting.
            else:  # spotted this by the time H and L etc were done. This was fixed for H/L to use DATASEC directly.
                xmin = 33
            retval.append(xmin)
            # xmax, ymin and ymax were all correctly taken from DATASEC
            retval.append(2080)
            retval.append(1)
            retval.append(4612)
            # the search area for E and O is/was [33:2080;1:4612], correctly, and same for H/L.
            retval.append(33)
            retval.append(2080)
            retval.append(1)
            retval.append(4612)

            with open(outfile, 'a') as ofile:
                # expnum, ra, dec, obs_end, mjd_middle, exptime (seconds)
                ofile.write('{} {} {} {} {} {} {} {} {} {} {} {} {}\n'.format(*retval))

        with open(outfile, 'a') as ofile:  # blank line between triplets
            ofile.write('\n')

            # copy up the file to VOSpace



#!python
from astropy import units
from astropy.coordinates import SkyCoord
from astropy.time import TimeDelta, Time
import numpy as np
from ossos.cameras import Camera
from ossos import mpc
import mp_ephem
from ossos.ephem_target import EphemTarget
import sys
import os
import argparse
import logging
import math
from copy import deepcopy
import ephem

cfht = ephem.Observer()
cfht.lat = 0.344
cfht.lon = -2.707
cfht.elevation = 4100
cfht.date = '2018/03/28 20:00:00'
sun = ephem.Sun()
fb = ephem.FixedBody()


def is_up(coordinate, current_time):
    """
    Given the position and time determin if the given target is up.

    @param coordinate: the J2000 location of the source
    @param current_time: The time of the observations
    @return: True/False
    """
    cfht.date = current_time.iso.replace('-', '/')
    cfht.horizon = math.radians(-7)
    sun.compute(cfht)
    sun_rise = Time(str(sun.rise_time).replace('/', '-'))
    sun_set = Time(str(sun.set_time).replace('/', '-'))

    if current_time < sun_set or current_time > sun_rise:
        return False

    fb._ra = coordinate.ra.radian
    fb._dec = coordinate.dec.radian
    cfht.horizon = math.radians(40)
    fb.compute(cfht)
    fb_rise_time = Time(str(fb.rise_time).replace('/', '-'))
    fb_set_time = Time(str(fb.set_time).replace('/', '-'))

    if (current_time > fb_set_time > fb_set_time or
       fb_rise_time > current_time > fb_set_time):
        return False
    return True


def create_ephemeris_file(name, camera, kbos, orbits, pointing_date, runid, ephem_format="CFHT API"):
    """

    @param name: name of the pointing
    @param camera: camera object that is this pointing
    @param kbos: list of kbos covered by this pointing
    @param orbits: orbits of all known kbos
    @param pointing_date: data of pointing associated with camera object
    @param runid:  The CFHT RunID that will be used to observer the target.
    @return: None
    """

    et = EphemTarget(name, ephem_format=ephem_format, runid=runid)
    # determine the mean motion of target KBOs in this field.
    field_kbos = []
    center_ra = 0
    center_dec = 0

    pointing_date = mpc.Time(pointing_date)
    start_date = pointing_date - TimeDelta(8.1 * units.day)
    end_date = start_date + TimeDelta(17 * units.day)
    time_step = TimeDelta(1.5 * units.hour)

    # Compute the mean position of KBOs in the field on current date.
    for kbo_name in kbos:
        kbo = orbits[kbo_name]
        kbo.predict(pointing_date)
        ra = kbo.coordinate.ra
        dec = kbo.coordinate.dec
        field_kbos.append(kbo)
        center_ra += ra.radian
        center_dec += dec.radian

    today = start_date
    while today < end_date:
        today += time_step
        mean_motion = (0, 0)
        max_mag = 0.0
        if len(field_kbos) > 0:
            current_ra = 0
            current_dec = 0
            for kbo in field_kbos:
                kbo.predict(today)
                max_mag = max(max_mag, kbo.r_mag)
                current_ra += kbo.coordinate.ra.radian
                current_dec += kbo.coordinate.dec.radian
            mean_motion = ((current_ra - center_ra) / len(field_kbos),
                           (current_dec - center_dec) / len(field_kbos))
        ra = camera.coordinate.ra.radian + mean_motion[0]
        dec = camera.coordinate.dec.radian + mean_motion[1]
        cc = SkyCoord(ra=ra,
                      dec=dec,
                      unit=(units.radian, units.radian),
                      obstime=today)
        if not is_up(cc, today):
            continue
        dt = pointing_date - today
        cc.dra = (mean_motion[0] * units.radian / dt.to(units.hour)).to(units.arcsec / units.hour).value * math.cos(dec)
        cc.ddec = (mean_motion[1] * units.radian / dt.to(units.hour)).to(units.arcsec / units.hour).value
        cc.mag = max_mag
        et.append(cc)

    et.save()
    return


def optimize(orbits, required, locations, tokens, camera_name="MEGACAM_40"):
    """

    @param orbits: a dictionary of orbits to optimize the point of.
    @param required: list of objects that MUST be observed.
    @param locations: locations of the objects at the start of observing period.
    @param tokens: List of tokens (object names) that we will try and cover.
    @return:
    """

    # Get the tokens that will be paged through in random order.
    token_order = np.random.permutation(required)
    optimal_pointings = {}
    covered = []  # the objects that have already been covered by a planned pointing.
    search_order = [0, ]
    # search_order = [22 + 4]
    # search_order.extend(range(len(Camera.names)))

    # For each required target find the pointing that will include the largest number of other required targets
    # and then tweak that specific pointing to include the maximum number of secondary targets.
    for token in token_order:
        if token in covered:
            continue
        logging.debug("Optimizing pointing for required target: {}".format(token))
        if token not in orbits:
            logging.error("No orbit available for: {}".format(token))
            continue
        obj = orbits[token]
        separations = obj.coordinate.separation(locations)
        possible_tokens = tokens[separations < 1.3 * units.degree]
        this_required = []
        for this_token in required:
            if this_token in possible_tokens:
                this_required.append(this_token)
        # This object is not inside the existing coverage.
        max_sources_in_pointing = 0
        best_coverage = []
        p = SkyCoord(obj.coordinate.ra,
                     obj.coordinate.dec)
        pointing = Camera(p, camera=camera_name)
        if len(possible_tokens) == 1:
            optimal_pointings[token] = pointing, possible_tokens
            logging.info(" {} is all alone!".format(token))
            continue

        logging.debug("examining possible optimizations")

        if len(this_required) == 1:
            best_coverage = [token]
        else:
            logging.debug("Maximizing inclusion of required targets")
            for idx in search_order:
                pointing.offset(index=idx)
                this_coverage = []
                for this_token in this_required:
                    if this_token in covered:
                        continue
                    this_obj = orbits[this_token]
                    if pointing.polygon.isInside(this_obj.coordinate.ra.degree, this_obj.coordinate.dec.degree):
                        this_coverage.append(this_token)

                if len(this_coverage) > max_sources_in_pointing:
                    max_sources_in_pointing = len(this_coverage)
                    best_coverage = this_coverage

        # best_pointing now contains the pointing that covers the largest number of required sources.
        # best_coverage is the list of sources (tokens) that this pointing covered.
        # now move around the best_pointing, keeping all the objects listed in best_coverage but maximizing the
        # number of other sources that get coverage (optimal_coverage)
        logging.info("{} pointing these {} required targets: {}".format(token, len(best_coverage), best_coverage))
        max_sources_in_pointing = 0
        optimal_coverage = []
        optimal_pointing = None
        for idx in search_order:
            pointing.offset(index=idx)
            exclude = False  # exclude this offset because it doesn't overlap one or more of the required sources.
            for this_token in best_coverage:
                this_obj = orbits[this_token]
                if not pointing.polygon.isInside(this_obj.coordinate.ra.degree, this_obj.coordinate.dec.degree):
                    exclude = True
                    break

            if exclude:
                continue

            # OK, this offset has all the required sources possible, how many extra sources do we get?
            this_coverage = []
            for this_token in possible_tokens:
                if this_token in covered:
                    continue
                this_obj = orbits[this_token]
                if pointing.polygon.isInside(this_obj.coordinate.ra.degree, this_obj.coordinate.dec.degree):
                    this_coverage.append(this_token)

            if len(this_coverage) > max_sources_in_pointing:
                max_sources_in_pointing = len(this_coverage)
                optimal_coverage = this_coverage
                optimal_pointing = idx

        # remove all sources covered by optimal_pointing from further consideration.
        sys.stdout.write("{} pointing covers: ".format(token))
        unique_coverage_list = []
        for this_token in optimal_coverage:
            if this_token not in covered:
                unique_coverage_list.append(this_token)
                sys.stdout.write(" {} ".format(this_token))
                covered.append(this_token)
        sys.stdout.write("\n")
        pointing.offset(index=optimal_pointing)
        optimal_pointings[token] = pointing, unique_coverage_list
        n = 0
        new_required = []
        for this_token in token_order:
            if this_token not in covered:
                new_required.append(this_token)
                n += 1
        logging.info("Remaining required targets: {}, targets covered: {}".format(n, len(covered)))
        required = new_required
    return optimal_pointings


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('date',
                        type=str,
                        help="ISO date string to optimize the pointings too")
    parser.add_argument('runid',
                        help="CFHT RUN ID")
    parser.add_argument('pointing_objects',
                        type=str,
                        help="Name of file that contains list of objects to try and observe.")
    parser.add_argument('required_objects',
                        type=str,
                        help="name of file that contains a list of required objects")
    parser.add_argument('--nattempts', type=int,
                        help="Number of random variations of pointings to try",
                        default=2)
    parser.add_argument('--camera', default="MEGACAM_40",
                        choices=list(Camera._geometry.keys()),
                        help="Name of camera")
    parser.add_argument('--ephem-format', default='CFHT_API',
                        choices=['CFHT_API', 'CFHT_ET', 'GEMINI_ET'])

    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO)

    pointing_date = args.date

    logging.info("Date for orbit predictions: {}".format(args.date))
    logging.info("Building pointings that include all targets in {}".format(args.required_objects))
    logging.info("Optimized to include as many objects in {} as possible".format(args.pointing_objects))

    required_objects = []
    required_filename = args.required_objects
    for required_object in open(required_filename):
        if '#' in required_object:
            continue
        required_objects.append(required_object.strip())

    pointings_filename = os.path.splitext(required_filename)[0] + "_pointings.txt"

    filenames = open(args.pointing_objects).readlines()

    orbits = {}
    tokens = []

    # load the orbits of all objects of interest.
    for filename in filenames:
        filename = filename.strip()
        token = os.path.splitext(os.path.basename(filename))[0]
        abg_filename = os.path.splitext(filename)[0] + ".abg"
        orbits[token] = mp_ephem.BKOrbit(None, ast_filename=filename, abg_file=abg_filename)
        tokens.append(token)

    # Turn the object locations at the time of interest into a SkyCoord numpy array.
    tokens = np.array(tokens)
    minimum_number_of_pointings = len(required_objects)

    best_pointing_list = []
    for attempt in range(args.nattempts):
        locations = []
        for token in tokens:
            orbits[token].predict(pointing_date)
            locations.append([orbits[token].coordinate.ra.degree, orbits[token].coordinate.dec.degree])
        locations = SkyCoord(locations, unit='degree')
        logging.info("Attempt : {} \n".format(attempt))
        pointings = optimize(orbits, required_objects, locations, tokens)
        if minimum_number_of_pointings >= len(pointings):
            minimum_number_of_pointings = len(pointings)
            best_pointing_list = deepcopy(pointings)

    with open(pointings_filename, 'w') as pobj:
        pobj.write("index {}\n".format(pointing_date))
        pointing_number = 0
        for token in best_pointing_list:
            pobj.write("{} {} {} {} # {}\n".format(pointing_number + 1,
                                                   "{}".format(token),
                                                   best_pointing_list[token][0].coord.to_string("hmsdms", sep=" "),
                                                   2000,
                                                   len(best_pointing_list[token][1])))
            pointing_number += 1
            create_ephemeris_file(token, best_pointing_list[token][0], best_pointing_list[token][1], orbits,
                                  pointing_date, args.runid, ephem_format=args.ephem_format.replace("_", " "))


if __name__ == '__main__':
    sys.exit(main())



__author__ = 'bannisterm'

from contextlib import contextmanager

# no. columns must match no. c's
# table has small font and is horizontal.
header = r"\begin{deluxetable}{ccccccccccccc}" + '\n' + \
         r"\tabletypesize{\scriptsize}" + '\n' + \
         r"\rotate" + '\n' + \
         r"\tablecolumns{13}" + '\n' + \
         r"\tablehead{\colhead{$m_{r}$} \vspace{-0.2cm} & " \
         r"\colhead{$\sigma$ $m_{r}$} & " \
         r"\colhead{Detectability} & " \
         r"\colhead{RA} & " \
         r"\colhead{Dec} & " \
         r"\colhead{a} & " \
         r"\colhead{e} & " \
         r"\colhead{i} & " \
         r"\colhead{r$_{H}$} & " \
         r"\colhead{H$_{r}$} & " \
         r"\colhead{MPC} & " \
         r"\colhead{Object} & " \
         r"\colhead{Status} \\" \
         r"\colhead{discovery} & " \
         r"\colhead{all obs} & " \
         r"\colhead{} & " \
         r"\colhead{discov.} & " \
         r"\colhead{discov.} & " \
         r"\colhead{(AU)} & " \
         r"\colhead{} & " \
         r"\colhead{($^{\circ}$)} & " \
         r"\colhead{(AU)} & " \
         r"\colhead{} & " \
         r"\colhead{design.} & " \
         r"\colhead{} & " \
         r"\colhead{} " \
         + r"}" \
         + "\n" \
         + "\startdata \n" \
         + r"\cutinhead{Centaurs}" + "\n"

footer = r"\enddata " + "\n" + \
         r"\tablecomments{$p:q$: object is in the $p:q$ resonance; I: the orbit classification is currently " \
         r"insecure; " \
         r"" \
         r"" \
         r"H: the human operator intervened to declare the orbit security status. " \
         r"$a, e, i$ are J2000 ecliptic barycentric coordinates, with uncertainties from the covariant matrix fit " \
         r"of \citet{Bernstein:2000p444}; full barycentric elements are available at \url{http://www.ossos-survey" \
         r".org/}." \
         r"The full heliocentric orbital elements are available in electronic form from the Minor Planet Center.} " "\n" + \
         "\end{deluxetable} \n"

extreme_header = r"\begin{deluxetable*}{lDDDDDDDD}" + '\n' + \
                 r"\tablecolumns{10}" + '\n' + \
                 r"\tablecaption{Heliocentric osculating orbital elements of known trans-Neptunian objects " \
                 r"with $q > 40$ and $a > 160$ au for epoch MJD 57600" + '\n' + \
                 r"\label{tab:orbits}"+ '\n' + \
                 r"}"+ '\n' + \
                 r"\tablehead{" \
                 r"\colhead{Name} & " \
                 r"\twocolhead{$q$} & " \
                 r"\twocolhead{$a$} & " \
                 r"\twocolhead{$e$} & " \
                 r"\twocolhead{$i$} & " \
                 r"\twocolhead{$\Omega$} & " \
                 r"\twocolhead{$\omega$} & " \
                 r"\twocolhead{Arc} & " \
                 r"\twocolhead{Discovery} \\ " \
                 r"\colhead{} & " \
                 r"\twocolhead{(au)} & " \
                 r"\twocolhead{(au)} & " \
                 r"\twocolhead{} & " \
                 r"\twocolhead{($^{\circ}$)} & " \
                 r"\twocolhead{($^{\circ}$)} & " \
                 r"\twocolhead{($^{\circ}$)} & " \
                 r"\twocolhead{(days)}  & " \
                 r"\twocolhead{survey} " \
                 r"}"+ '\n' + \
                 r"\startdata "+ '\n'

extreme_footer = r"\enddata"+ '\n' + \
                 r"\tablecomments{Orbital elements retrieved from JPL Horizons on 6 Oct 2016."+ '\n' + \
                 r"$\Omega$ is the longitude of ascending node, $\omega$ the argument of perihelion. " \
                 r"The uncertainties are the $1\sigma$ estimates based on the covariance matrix at the best-fit orbit.}" \
                 r"\end{deluxetable*} "+ '\n'



@contextmanager
def deluxe_table_formatter(outfile):
    ofile = open(outfile, 'w')
    ofile.write(header)

    yield ofile

    ofile.write(footer)
    ofile.close()

@contextmanager
def extreme_tno_table_formatter(outfile):
    ofile = open(outfile, 'w')
    ofile.write(extreme_header)

    yield ofile

    ofile.write(extreme_footer)
    ofile.close()


#!python 
################################################################################
##                                                                            ##
## Copyright 2013 by its authors                                              ##
## See COPYING, AUTHORS                                                       ##
##                                                                            ##
## This file is part of OSSOS Moving Object Pipeline (OSSOS-MOP)              ##
##                                                                            ##
##    OSSOS-MOP is free software: you can redistribute it and/or modify       ##
##    it under the terms of the GNU General Public License as published by    ##
##    the Free Software Foundation, either version 3 of the License, or       ##
##    (at your option) any later version.                                     ##
##                                                                            ##
##    OSSOS-MOP is distributed in the hope that it will be useful,            ##
##    but WITHOUT ANY WARRANTY; without even the implied warranty of          ##
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           ##
##    GNU General Public License for more details.                            ##
##                                                                            ##
##    You should have received a copy of the GNU General Public License       ##
##    along with OSSOS-MOP.  If not, see <http://www.gnu.org/licenses/>.      ##
##                                                                            ##
################################################################################
"""Run the OSSOS makepsf proceedure"""

import argparse
import os
import sys
import logging
from ossos import storage
from ossos import util

task = 'mkpsf'
dependency = 'mk_mopheader'


def run(expnum, ccd, version, dry_run=False, prefix="", force=False):
    """Run the OSSOS jmpmakepsf script.

    """

    message = storage.SUCCESS
    if storage.get_status(task, prefix, expnum, version=version, ccd=ccd) and not force:
        logging.info("{} completed successfully for {} {} {} {}".format(task, prefix, expnum, version, ccd))
        return

    with storage.LoggingManager(task, prefix, expnum, ccd, version, dry_run):
        try:
            if not storage.get_status(dependency, prefix, expnum, version, ccd=ccd):
                raise IOError("{} not yet run for {}".format(dependency, expnum))

            # confirm destination directory exists.
            destdir = os.path.dirname(
                storage.dbimages_uri(expnum, ccd, prefix=prefix, version=version, ext='fits'))
            if not dry_run:
                storage.mkdir(destdir)

            # get image from the vospace storage area
            logging.info("Getting fits image from VOSpace")
            filename = storage.get_image(expnum, ccd, version=version, prefix=prefix)

            # get mopheader from the vospace storage area
            logging.info("Getting mopheader from VOSpace")
            mopheader_filename = storage.get_file(expnum, ccd, version=version, prefix=prefix, ext='mopheader')

            # run mkpsf process
            logging.info("Running mkpsf on %s %d" % (expnum, ccd))
            logging.info(util.exec_prog(['jmpmakepsf.csh',
                                         './',
                                         filename,
                                         'yes', 'yes']))
            
            if dry_run:
                return

            # place the results into VOSpace
            basename = os.path.splitext(filename)[0]

            for ext in ('mopheader', 'psf.fits',
                        'zeropoint.used', 'apcor', 'fwhm', 'phot'):
                dest = storage.dbimages_uri(expnum, ccd, prefix=prefix, version=version, ext=ext)
                source = basename + "." + str(ext)
                count = 0
                with open(source, 'r'):
                  while True:
                    count += 1
                    try:
                        logging.info("Attempt {} to copy {} -> {}".format(count, source, dest))
                        storage.copy(source, dest)
                        break
                    except Exception as ex:
                        if count > 10:
                            raise ex

            # set some data parameters associated with the image, determined in this step.
            storage.set_status('fwhm', prefix, expnum, version=version, ccd=ccd, status=str(storage.get_fwhm(
                expnum, ccd=ccd, prefix=prefix, version=version)))
            storage.set_status('zeropoint', prefix, expnum, version=version, ccd=ccd,
                               status=str(storage.get_zeropoint(
                                   expnum, ccd=ccd, prefix=prefix, version=version)))
            logging.info(message)
        except Exception as e:
            message = str(e)
            logging.error(message)
            
        storage.set_status(task, prefix, expnum, version, ccd=ccd, status=message)

    return


def main():


    parser = argparse.ArgumentParser(
        description='Run makepsf chunk of the OSSOS pipeline')

    parser.add_argument('--ccd', '-c',
                        action='store',
                        type=int,
                        dest='ccd',
                        default=None,
                        help='which ccd to process, default is all')
    parser.add_argument("--dbimages",
                        action="store",
                        default="vos:OSSOS/dbimages",
                        help='vospace dbimages containerNode')
    parser.add_argument("expnum",
                        type=int,
                        nargs='+',
                        help="expnum(s) to process")
    parser.add_argument("--dry-run",
                        action="store_true",
                        help="DRY RUN, don't copy results to VOSpace, implies --force")

    parser.add_argument("--fk", action="store_true", help="Run fk images")

    parser.add_argument("--type", "-t", choices=['o', 'p', 's'],
                        help="which type of image: o-RAW, p-ELIXIR, s-SCRAMBLE", default='p')
    parser.add_argument("--verbose", "-v",
                        action="store_true")
    parser.add_argument("--force", default=False,
                        action="store_true")
    parser.add_argument("--debug", "-d",
                        action="store_true")

    cmd_line = " ".join(sys.argv)
    args = parser.parse_args()

    util.set_logger(args)
    logging.info("Started {}".format(cmd_line))

    prefix = (args.fk and 'fk') or ''

    storage.DBIMAGES = args.dbimages

    exit_code = 0
    for expnum in args.expnum:
        if args.ccd is None:
           if int(expnum) < 1785619:
               # Last exposures with 36 CCD Megaprime
               ccdlist = list(range(0,36))
           else:
               # First exposrues with 40 CCD Megaprime
               ccdlist = list(range(0, 40))
        else:
           ccdlist = [args.ccd]
        for ccd in ccdlist:
            run(expnum, ccd, args.type, args.dry_run, prefix, args.force)
    return exit_code

if __name__ == '__main__':
    sys.exit(main())



import argparse
import json

import numpy
from astropy.coordinates import SkyCoord


class Program(object):
    def __init__(self, runid="16BP06", pi_login="gladman"):
        self.config = {"runid": runid,
                       "pi_login": pi_login,
                       "program_configuration": {"targets": [],
                                                 "observing_blocks": [],
                                                 "observing_groups": []
                                                 }}

    def add_target(self, target):
        self.config["program_configuration"]["targets"].append(target)

    def add_observing_block(self, observing_block):
        self.config["program_configuration"]["observing_blocks"].append(observing_block)

    def add_observing_group(self, observing_group):
        self.config["program_configuration"]["observing_groups"].append(observing_group)


class Target(object):
    def __init__(self, filename=None):
        self.config = json.load(open(filename))

    @property
    def token(self):
        return self.config["identifier"]["client_token"]

    @property
    def mag(self):
        return self.config["moving_target"]["ephemeris_points"][0]["mag"]

    @property
    def coordinate(self):
        return SkyCoord(self.config["moving_target"]["ephemeris_points"][0]["coordinate"]["ra"],
                        self.config["moving_target"]["ephemeris_points"][0]["coordinate"]["dec"],
                        unit='degree')


class ObservingBlock(object):
    def __init__(self, client_token, target_token):
        self.config = {"identifier": {"client_token": client_token},
                       "target_identifier": {"client_token": target_token},
                       "constraint_identifiers": [{"server_token": "C1"}],
                       "instrument_config_identifiers": [{"server_token": "I1"}]}

    @property
    def token(self):
        return self.config["identifier"]["client_token"]


class ObservingGroup(object):
    def __init__(self, client_token):
        self.config = {"identifier": {"client_token": client_token},
                       "observing_block_identifiers": []}

    def add_ob(self, client_token):
        self.config["observing_block_identifiers"].append({"client_token": client_token})


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('pi_login')
    parser.add_argument('qrunid')
    parser.add_argument('runid')
    parser.add_argument('targets', nargs='+')
    parser.add_argument('--og-idx-start', default=0)
    args = parser.parse_args()

    # Break the targets into OBs based on their max mag of source in pointing.
    cuts = numpy.array([23.0, 23.5, 24.0, 24.5, 25.0, 25.5, 26.0, 30.0])
    IC_exptimes = [50, 100, 200, 300, 400, 500, 600, 700]

    program = Program(runid=args.runid, pi_login=args.pi_login)
    ob_tokens = []
    mags = {}
    ob_coordinate = {}
    for filename in args.targets:
        target = Target(filename)
        program.add_target(target.config)
        ob_token = "OB-{}-{}".format(target.token, target.mag)
        ob = ObservingBlock(ob_token, target.token)
        idx = (target.mag > cuts).sum() + 4
        idx = 2
        ob.config["instrument_config_identifiers"] = [{"server_token": "I{}".format(idx)}]
        program.add_observing_block(ob.config)
        ob_tokens.append(ob_token)
        mags[ob_token] = target.mag
        ob_coordinate[ob_token] = target.coordinate

    sf = lambda x, y: (x.ra < y.ra) - (y.ra < x.ra)
    order_tokens = sorted(ob_coordinate, cmp=sf, key=ob_coordinate.get)
    total_itime = 0
    ogs = {}
    scheduled = {}
    og_idx = int(args.og_idx_start)
    if False:
        # while len(scheduled) < len(ob_tokens):
        og_idx += 1
        og_token = "OG_{}_{}_{}_{}".format(args.runid, args.qrunid, og_idx, 0)
        sys.stdout.write("{}: ".format(og_token))
        og = ObservingGroup(og_token)
        og_coord = None
        og_itime = 0
        for ob_token in order_tokens:
            if ob_token not in scheduled:
                if og_coord is None:
                    og_coord = ob_coordinate[ob_token]
                if ob_coordinate[ob_token].separation(og_coord) > 30 * units.degree:
                    continue
                og.add_ob(ob_token)
                scheduled[ob_token] = True
                sys.stdout.write("{} ".format(ob_token))
                sys.stdout.flush()
                idx = (mags[ob_token] > cuts).sum()
                idx = 1
                print(ob_token, mags[ob_token], idx + 4)
                og_itime += IC_exptimes[idx] + 40
                if og_itime > 3000.0:
                    break
                break
        total_itime += og_itime
        sys.stdout.write(" {}s \n".format(og_itime))
        program.add_observing_group(og.config)
        nrepeats = 0
        for repeat in range(nrepeats):
            total_itime += og_itime
            og_token = "OG_{}_{}_{}_{}".format(args.runid, args.qrunid, og_idx, repeat + 1)
            og = copy.deepcopy(og)
            og.config["identifier"]["client_token"] = og_token
            program.add_observing_group(og.config)
    # print "Total I-Time: {} hrs".format(total_itime/3600.)
    json.dump(program.config, open('program.json', 'w'), indent=4, sort_keys=True)



#!python 
################################################################################
##                                                                            ##
## Copyright 2013 by its authors                                              ##
## See COPYING, AUTHORS                                                       ##
##                                                                            ##
## This file is part of OSSOS Moving Object Pipeline (OSSOS-MOP)              ##
##                                                                            ##
##    OSSOS-MOP is free software: you can redistribute it and/or modify       ##
##    it under the terms of the GNU General Public License as published by    ##
##    the Free Software Foundation, either version 3 of the License, or       ##
##    (at your option) any later version.                                     ##
##                                                                            ##
##    OSSOS-MOP is distributed in the hope that it will be useful,            ##
##    but WITHOUT ANY WARRANTY; without even the implied warranty of          ##
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           ##
##    GNU General Public License for more details.                            ##
##                                                                            ##
##    You should have received a copy of the GNU General Public License       ##
##    along with OSSOS-MOP.  If not, see <http://www.gnu.org/licenses/>.      ##
##                                                                            ##
################################################################################
"""plant synthetic moving objects into a set of observations.

prior to planting, the headers of the objects may be swapped around."""

import argparse
import logging
import os
import sys

from astropy.io import fits

from ossos import storage
from ossos import util

dependency = "update_header"
task = 'scramble'


def scramble(expnums, ccd, version='p', dry_run=False, force=False, prefix=''):
    """
    run the plant script on this combination of exposures

    @param expnums: list of exposure numbers to scramble the time on
    @param ccd:  which CCD in (assumes this is a CFHT MegaCam MEF)
    @param version: should we scramble the 'p' or 'o' images?
    @param dry_run: if dry run then don't save back to VOSpace.
    @param force: if true then create scramble set, even if already exists.
    @param prefix: a string that will be pre-pended to the EXPNUM to get the filename, sometimes 'fk'.
    @return: None
    """

    # Get a list of the MJD values and then write a re-ordering of those into files with 's'
    # as their type instead of 'p' or 'o'
    mjds = []
    fobjs = []
    message = storage.SUCCESS
    if not (force or dry_run) and storage.get_status(task, prefix, expnums[0],
                                                     version='s', ccd=ccd):
        logging.info("{} recorded as complete for {} ccd {}".format(task,
                                                                    expnums, ccd))
        return

    with storage.LoggingManager(task, prefix, expnums[0], ccd, version):
        try:
            for expnum in expnums:
                filename = storage.get_image(expnum, ccd=ccd, version=version)
                fobjs.append(fits.open(filename))
                # Pull out values to replace in headers.. must pull them
                # as otherwise we get pointers...
                mjds.append(fobjs[-1][0].header['MJD-OBS'])

            order = [0, 2, 1]
            for idx in range(len(fobjs)):
                logging.info("Flipping %d to %d" % (fobjs[idx][0].header['EXPNUM'],
                                                    expnums[order[idx]]))
                fobjs[idx][0].header['EXPNUM'] = expnums[order[idx]]
                fobjs[idx][0].header['MJD-OBS'] = mjds[order[idx]]
                uri = storage.get_uri(expnums[order[idx]],
                                      ccd=ccd,
                                      version='s',
                                      ext='fits')
                scramble_file_name = os.path.basename(uri)
                if os.access(scramble_file_name, os.F_OK):
                    os.unlink(scramble_file_name)
                fobjs[idx].writeto(scramble_file_name)
                if not dry_run:
                    storage.copy(scramble_file_name, uri)
            logging.info(message)
        except Exception as ex:
            message = str(ex)
            logging.error(message)

        if not dry_run:
            storage.set_status(task, prefix, expnum, version, ccd, status=message)

    return


def main():
    """Run the script."""

    parser = argparse.ArgumentParser(description='time scramble the input images by tweaking their headers.')
    parser.add_argument('--ccd',
                        action='store',
                        type=int,
                        default=None,
                        help='which ccd to process')
    parser.add_argument("--dbimages",
                        action="store",
                        default="vos:OSSOS/dbimages",
                        help='vospace dbimages containerNode')
    parser.add_argument("expnums",
                        type=int,
                        nargs=3,
                        help="expnums to scramble")
    parser.add_argument("--nosort",
                        action='store_true',
                        default=False,
                        help="do not sort before processing")
    parser.add_argument("--type",
                        action='store',
                        default='p',
                        choices=['p', 'o'],
                        help='which type of image')
    parser.add_argument("--verbose", "-v",
                        action="store_true")
    parser.add_argument("--debug", '-d',
                        action='store_true')
    parser.add_argument("--dry-run", action="store_true", help="Do not copy back to VOSpace, implies --force")
    parser.add_argument("--force", action='store_true')

    cmd_line = " ".join(sys.argv)
    args = parser.parse_args()
    util.set_logger(args)
    logging.info('Starting {}'.format(cmd_line))

    storage.DBIMAGES = args.dbimages
    prefix = ''
    version = args.type

    if not args.nosort:
        args.expnums.sort()

    expnums = args.expnums
    if args.ccd is None:
        if int(expnums[0]) < 1785619:
            # Last exposures with 36 CCD Megaprime
            ccdlist = list(range(0, 36))
        else:
            # First exposrues with 40 CCD Megaprime
            ccdlist = list(range(0, 40))
    else:
        ccdlist = [args.ccd]
    for ccd in ccdlist:
        # check if scramble image alraedy made for this ccd
        scramble(expnums=expnums, ccd=ccd, version=version, dry_run=args.dry_run, prefix=prefix)


if __name__ == '__main__':
    sys.exit(main())
